import os 
import gdown 
from pathlib import Path 
import platform
import subprocess
from dart_physics.dirs import * 

CUR_PATH = Path(__file__).parent.absolute()


GDRIVE_URL_MAP = {
    'yellow_drill': {
        'usdz': 'https://drive.google.com/file/d/1oV-Dyjdd22SljjD9UoA1Rv0K2s-DTJrE/view?usp=sharing', 
        'obj': 'https://drive.google.com/file/d/166iY57usZZrc0PXBiwBXEP599KL7jrKC/view?usp=sharing',
        'gltf': 'https://drive.google.com/file/d/1luhLk-1GZLs3Di_-axxy0UEMrC0vLDTQ/view?usp=sharing',
    },
    'yellow_drill_battery': {
        'obj': 'https://drive.google.com/file/d/1DC0BFLaLKW18gf0OXwmnpM7gYzBnuABI/view?usp=sharing', 
        'usdz': 'https://drive.google.com/file/d/1aWWi0kGzBgyKfJe1vunFyfjLIzXISOL1/view?usp=sharing', 
        'gltf': 'https://drive.google.com/file/d/1Oar0KiJ2S4KKt0TvcFlU3jBlCZA9y_Sy/view?usp=sharing',
    },
    'tongue': {
        'obj': 'https://drive.google.com/file/d/1G-zvUMpMewEihWbS0j36tX47uN3jooRg/view?usp=sharing', 
        'usdz': 'https://drive.google.com/file/d/17qhCX6guWPo5R8rxDKZG1NSVgDUBd_4e/view?usp=sharing',
        'gltf': 'https://drive.google.com/file/d/1abLOreAsmY-HivZaNBaot85uJ5uI2Idn/view?usp=sharing',
    },
    'red_marker': {
        'obj': 'https://drive.google.com/file/d/17u1t1BpybIsMDBR-o9H0EgPjpZYXrKgi/view?usp=sharing', 
        'usdz': 'https://drive.google.com/file/d/1Gtnlt-2FLk4LfvoCrqnTrpWfqXvlb56n/view?usp=sharing',
        'gltf': 'https://drive.google.com/file/d/1pBKIQE_B-uk0xvsRy5dIpNpMmyKUvx6e/view?usp=sharing', 
    },
    'red_ladle': {
        'obj':  'https://drive.google.com/file/d/1IC4TOuJwdIUJtyxsq3PXENs9_5ONLJ4i/view?usp=sharing',
        'usdz': 'https://drive.google.com/file/d/17CjiafFC7q08Q-NXgjfK3TV5Fik7QQ4k/view?usp=sharing', 
        'gltf': 'https://drive.google.com/file/d/1Haio-gOebVRqhMlKxhjps5IVOg3WPB0U/view?usp=sharing',
    },
    'milk_tea': {
        'obj': 'https://drive.google.com/file/d/1swvaJUwYtow07fTuGv8l19amZrYmYcgS/view?usp=sharing', 
        'usdz': 'https://drive.google.com/file/d/1LGbpiLrrPRakes5qsHBfhxHpI2tDCaSn/view?usp=sharing',
        'gltf': 'https://drive.google.com/file/d/1GH-yRxtR3M27jGUVDHGp6RxrdW0NAYVh/view?usp=sharing', 
    },
    'hanger': {
        'obj': 'https://drive.google.com/file/d/1eSiey-7sXlZpadDeWojKfVLwPEuC3ovN/view?usp=sharing', 
        'usdz': 'https://drive.google.com/file/d/1uEtXlncYOEq6MUVWsq7rqKJkVAq3XdIA/view?usp=sharing', 
        'gltf': 'https://drive.google.com/file/d/1LPYmCcSPAu66ELZPlwVSJBJBBVj7F3hN/view?usp=sharing',
    }, 
    'eraser': {
        'obj': 'https://drive.google.com/file/d/16yWjL55qxdgNfy7ZoJlldGZYcOuCdb0h/view?usp=sharing', 
        'usdz': 'https://drive.google.com/file/d/1zSN8hCHuszdUR3uucaYrF1rjnh2jK46n/view?usp=sharing',
        'gltf': 'https://drive.google.com/file/d/1dcD8vkPrp-7YRDCD4VtaoiLjqCbCjC2T/view?usp=sharing',
    },
    'emergency_kit': {
        'obj': 'https://drive.google.com/file/d/177-xepqmCnXkoiglszyk40B5m6q7m6bs/view?usp=sharing',
        'usdz': 'https://drive.google.com/file/d/1Cd2y5B3LiX_FTrCetECtRkj6ZgT-KGKI/view?usp=sharing', 
        'gltf': 'https://drive.google.com/file/d/1ETphpoipqnlujoQYTbVKsT9B3UOxaWoC/view?usp=sharing',
    },
    'drill': {
        'obj': 'https://drive.google.com/file/d/1-DpQaTe8IvcSDjXnom_XmkKdGSBxrNhW/view?usp=sharing',
        'usdz': 'https://drive.google.com/file/d/1RuSbBjj3n54uLiDI9xVcEhcLKaWLv54C/view?usp=sharing',
        'gltf': 'https://drive.google.com/file/d/1DHMhwWPrgrwu7wKSyEgVhPQRKdVWF1wm/view?usp=sharing',
    },
    'cleaner2': {
        'obj': 'https://drive.google.com/file/d/18MqSwqxC5v0nnjSQ760o1r_wxUxYbPSo/view?usp=sharing', 
        'usdz': 'https://drive.google.com/file/d/1u6Q9GyeBgWbyrIybg87ntQXBR3NUy_W-/view?usp=sharing',
        'gltf': 'https://drive.google.com/file/d/15_G9RsS2xR32dWuOth4wfMqo1lV0Xz8u/view?usp=sharing',
    },
    'cleaner': {
        'obj': 'https://drive.google.com/file/d/1fnIXmW71oBPvltHqcUE0fH4jwqEpqXgQ/view?usp=sharing', 
        'usdz': 'https://drive.google.com/file/d/1eMSoWYhN9SiN5CXQu3kOKh9JTjL-E82N/view?usp=sharing',
        'gltf': 'https://drive.google.com/file/d/1kYVERsbCy6SFpPBofaGToZFkzXgP3Jyl/view?usp=sharing',
    },
    'brown_basket': {
        'obj': 'https://drive.google.com/file/d/1ZG3u9DlWDhnI39Iac2G4eI9ViqC9ZAY8/view?usp=sharing', 
        'usdz': 'https://drive.google.com/file/d/1svaDdFu2_Nkctt00tSstCIjzniHevsFb/view?usp=sharing', 
        'gltf': 'https://drive.google.com/file/d/1CE4EeCOtd3piAe2w76lJ-jQhXfu6EmQX/view?usp=sharing',
    },
    'bluebottle_cup': {
        'obj': 'https://drive.google.com/file/d/1Nchl_y1Pb0tz36RGgVgDPakSxv6NB-UZ/view?usp=sharing', 
        'usdz': 'https://drive.google.com/file/d/1GSI5aZu9HnBDAYebuhQork_q4Eqn9TUd/view?usp=sharing', 
        'gltf': 'https://drive.google.com/file/d/1bzznDRiITn7OgvscKnqT2GE2i6bgdX4x/view?usp=sharing',
    },
    'blue_marker': {
        'obj': 'https://drive.google.com/file/d/1rjJNKSRvIuMLuhiE3j3dhcWPVxJr_E9T/view?usp=sharing', 
        'usdz': 'https://drive.google.com/file/d/1Sc1EI3AdNJC1hRoOZV15pKzEAuR3MX93/view?usp=sharing',
        'gltf': 'https://drive.google.com/file/d/1N5qh1ILddBs32KhBRNR2953e7t8qguxk/view?usp=sharing',
    },
    'bidex_table': {
        '0down': 'https://drive.google.com/file/d/1chd7S3Rx9Lf8N3QeUWAM0aTMQr22Y-dv/view?usp=sharing',
        '100down': 'https://drive.google.com/file/d/1_EHNWGFolg33HmS9V6A94ncvMQiLdqGW/view?usp=sharing',
    },
    'cambridge_mug': {
        'usdz': 'https://drive.google.com/file/d/1QSpbPFkB44w11HKk0JNfiW8oE_VlOxja/view?usp=sharing',
        'obj': 'https://drive.google.com/file/d/1CpFWNr0p4d6vZS8NZfenaC--elK4bhns/view?usp=sharing', 
        'gltf': 'https://drive.google.com/file/d/1hQUXvqkBMlC7HXIhUvtzVoYtEFwkIuoI/view?usp=sharing',
    }, 
    'tesla_battery_slot': {
        'usdz': 'https://drive.google.com/file/d/16o9vfsDnQJ8UcaAkhejUPfDQxzSP8ekJ/view?usp=sharing', 
        'obj': 'https://drive.google.com/file/d/1810oOAxD_7uq8hlTYj2XhBB53oQVqgEO/view?usp=sharing', 
    }, 
    'cylinder': {
        'usdz': 'https://drive.google.com/file/d/1uGFmQrmC77CTEsu_QT1IPoabSBUKzLNN/view?usp=sharing', 
        'obj': 'https://drive.google.com/file/d/1R7d-4sdT37sKKxsynBmmszNXLvKr3IQI/view?usp=sharing', 
    },

    'screw_18402_1_final': {
        'gltf': 'https://drive.google.com/file/d/1dBBCbMU8Sy_tlIr_VBH2lLvlh9pAb_R1/view?usp=sharing', 
    },
    'screw_18402_0_final': {
        'gltf': 'https://drive.google.com/file/d/1V6Diufo42uMe8XS_syJtE4euuWoAsDod/view?usp=sharing', 
    },
    'screw_18402_0B_final': {
        'gltf': 'https://drive.google.com/file/d/1jgxY8LgyYZW0HcgHtVDAg-hdyEEinSTD/view?usp=sharing', 
    },
    'screw_18402_1B_final': {
        'gltf': 'https://drive.google.com/file/d/1hIAdjwgZpXqFAQGrBK5qiywt5QcFHWfj/view?usp=sharing', 
    },
    '00112_0L_final': {
        'gltf': 'https://drive.google.com/file/d/1cLRcWmsjp9fu9qREu6pcygl9F2h_NWZF/view?usp=sharing',
    },
    '00112_1L_final' : {
        'gltf': 'https://drive.google.com/file/d/1MQqKxgE7XkNr_xwqVliynWFGypxod6US/view?usp=sharing',
    },

    '00015_0L_final': {
        'gltf': 'https://drive.google.com/file/d/1vrMGSN0f5lhxSd2iTunqawP9RnHK6wzj/view?usp=sharing',
    },
    '00015_1L_final': {
        'gltf': 'https://drive.google.com/file/d/1SSkfAcm61IYZjOOmxho7pEE-QiKaQEap/view?usp=sharing',
    },

    '00470_0L_final': {
        'gltf': 'https://drive.google.com/file/d/1aP12xpkRTGsa9n5grAbD8mkjejlblbsT/view?usp=sharing',
    },
    '00470_1L_final': {
        'gltf': 'https://drive.google.com/file/d/1O0R3IdlM0wVJC_kT2zkk0R2iKBc_xZPX/view?usp=sharing',
    },

    '19133_0L_final': {
        'gltf': 'https://drive.google.com/file/d/1MLp0ytrj50b-ezsmhEjPq_o1dJsO94QT/view?usp=sharing',
    },
    '19133_1L_final': {
        'gltf': 'https://drive.google.com/file/d/18KBvmFzhXffJDdKPKRVFVOem08FcfopG/view?usp=sharing',
    },

    'keylight_0L_final': {
        'gltf': 'https://drive.google.com/file/d/1_giV82Or6sYZYjlGMRrHQI8tA2wL524F/view?usp=sharing',
    },
    'keylight_1L_final': {
        'gltf': 'https://drive.google.com/file/d/18R8VipHXtqDVvQ7bT5NQ0iiKYkN4cxQ6/view?usp=sharing',
    },

    'aj_0L_final': {
        'gltf': 'https://drive.google.com/file/d/1Wk64Dl8Dtm8Ed07pw73F5u-aOQ_7M4jl/view?usp=sharing',
    },
    'aj_1L_final': {
        'gltf': 'https://drive.google.com/file/d/1aPEYHgAuz4cpLCgCcA2H6cdzVjmtIL0V/view?usp=sharing',
    },

    '00932_0L_final': {
        'gltf': 'https://drive.google.com/file/d/1zbgiXwoaCJVHQLnZDkfVhm3TqyG0BoC1/view?usp=sharing',
    },
    '00932_1L_final': {
        'gltf': 'https://drive.google.com/file/d/1VUHfNYR3AivPoeo4Xk9-feUSal8rR9Nb/view?usp=sharing',
    },

    '6-32-screw_final': {
        'gltf': 'https://drive.google.com/file/d/1YRnK6gBtv_p1xiWGgMN44NGW5NbJlrT2/view?usp=sharing',
    },
    '6-32-screwdriver_final': {
        'gltf': 'https://drive.google.com/file/d/1ARiZBn2CQyAojD9wybf5gEnNv0_z5NYd/view?usp=sharing',
    },
    '6-32-screwmount_final': {
        'gltf': 'https://drive.google.com/file/d/10MBqVOM8fT3Np1o-1N48mdzxoKFBEpzI/view?usp=sharing',
    },

    '18-8-bolt_final': {
        'gltf': 'https://drive.google.com/file/d/1ut7Tdn668Qzpn7ef762Y1nlKoCuai5gO/view?usp=sharing',
    },
    '18-8-nut_final': {
        'gltf': 'https://drive.google.com/file/d/1P5SPcqRyQjaH37VaO5-9gHVd4qc7s_FU/view?usp=sharing',
    },

    'power-plug_final': {
        'gltf': 'https://drive.google.com/file/d/1C4hyx3DJzjaKcXWEQNMHN-bTbr514-yB/view?usp=sharing',
    },
    'power-socket_final': {
        'gltf': 'https://drive.google.com/file/d/142uyHdq-C4cKBYi8VSkQy2UeHUFFzBMT/view?usp=sharing',
    },
    'toyA_base': {
        'obj': 'https://drive.google.com/file/d/1QeUbD6OfQkHPWbqHZ9Io5aQeHkiE-csW/view?usp=sharing', 
    },
    'toyA_four': { 
        'obj': 'https://drive.google.com/file/d/1ECyuWum3SoWm89pelEsX1sJor8MQ5EiN/view?usp=sharing', 
    },
    'toyA_four_CAD': { 
        'obj': 'https://drive.google.com/file/d/1F2mHM-F-_XTYxfnW-OTVxvS_Ow6BRlHg/view?usp=sharing', 
    },
    'toyA_one_CAD': { 
        'obj': 'https://drive.google.com/file/d/1n8SEiy1i_-LKsvg68os1Gdscq5qVCrfY/view?usp=sharing', 
    },
    'muggood': {
        'obj': 'https://drive.google.com/file/d/1eowHU0_huBKvJDwQiy9W2jtMiRFW0LsN/view?usp=sharing',
    }
}

def download_mesh(mesh_name, mesh_type = 'obj', save_dir = ASSET_STORE_DIR): 

    if not os.path.exists(save_dir): 
        os.makedirs(save_dir)


    if mesh_type == 'obj':


        if os.path.exists(f'{save_dir}/{mesh_name}/{mesh_name}.obj'):
            print(f'{mesh_name}.obj already exists in {save_dir}')
            return f'{save_dir}/{mesh_name}'

        gdrive_url = GDRIVE_URL_MAP[mesh_name][mesh_type]

        gdown.download(gdrive_url, f'{save_dir}/{mesh_name}_obj.zip', fuzzy = True)

        # check if unzip is installed
        if not check_unzip_installed():
            raise RuntimeError("Unzip is not installed. Please install unzip command line tool.")
                
        # unzip the downloaded file
        os.makedirs(f'{save_dir}/{mesh_name}', exist_ok = True)
        os.system(f'unzip -o {save_dir}/{mesh_name}_obj.zip -d {save_dir}/{mesh_name}')
        if platform.system() != "Windows":
            os.system(f'chmod -R 755 {save_dir}/{mesh_name}')  # Sets read, write, and execute permissions for user, and read and execute for group and others
            os.system(f'rm {save_dir}/{mesh_name}_obj.zip')
        else:
            os.remove(f'{save_dir}/{mesh_name}_obj.zip')
        
    elif mesh_type == 'usdz':
        if os.path.exists(f'{save_dir}/{mesh_name}.usdz'):
            print(f'{mesh_name}.usdz already exists in {save_dir}')
            return
        
        gdrive_url = GDRIVE_URL_MAP[mesh_name][mesh_type]

        gdown.download(gdrive_url, f'{save_dir}/{mesh_name}.usdz', fuzzy = True)
    elif mesh_type == 'gltf':
        if os.path.exists(f'{save_dir}/{mesh_name}.gltf'):
            print(f'{mesh_name}.gltf already exists in {save_dir}')
            return
        
        gdrive_url = GDRIVE_URL_MAP[mesh_name][mesh_type]

        gdown.download(gdrive_url, f'{save_dir}/{mesh_name}.gltf', fuzzy = True)

    print(f'{mesh_name}.{mesh_type} downloaded to {save_dir}')

    return f'{save_dir}/{mesh_name}'


def download_table(table_type = '100down', save_dir = ASSET_STORE_DIR):

    gdrive_url = GDRIVE_URL_MAP['bidex_table'][table_type]
    if table_type == '0down': 
        stl_name = 'table.stl'
    else:
        stl_name = f'table_{table_type}.stl'

    if os.path.exists(f'{save_dir}/{stl_name}'):
        print(f'{stl_name} already exists in {save_dir}')
        return

    gdown.download(gdrive_url, f'{save_dir}/{stl_name}', fuzzy = True)



def create_urdf(mesh_name, save_dir = ASSET_STORE_DIR):
    # Create the URDF file path
    urdf_file_path = os.path.join(save_dir, f'{mesh_name}.urdf')

    # Get the mesh file path
    mesh_file_path = os.path.join(save_dir, mesh_name, f'{mesh_name}.obj')
    material_file_path = os.path.join(save_dir, mesh_name, f'baked_mesh_tex0.png')
    mesh_exists = os.path.exists(material_file_path)

    # Create the URDF content
    urdf_content = f"""<?xml version="1.0"?>
<robot name="{mesh_name}">
<link name="base_link">
    <visual>
    <geometry>
        <mesh filename="{mesh_file_path}" scale="1 1 1"/>
    </geometry>
    <material name="material">
        <texture filename="{material_file_path}"/>
    </material>
    </visual>
    <collision>
    <geometry>
        <mesh filename="{mesh_file_path}" scale="1 1 1"/>
    </geometry>
    </collision>
</link>
</robot>
"""
    
    no_mesh_urdf_content = f"""<?xml version="1.0"?>
<robot name="{mesh_name}">
<link name="base_link">
    <visual>
    <geometry>
        <mesh filename="{mesh_file_path}" scale="1 1 1"/>
    </geometry>
    </visual>
    <collision>
    <geometry>
        <mesh filename="{mesh_file_path}" scale="1 1 1"/>
    </geometry>
    </collision>
</link> 
</robot>
"""

    # Write the URDF content to the file
    with open(urdf_file_path, 'w') as urdf_file:
        if mesh_exists:
            urdf_file.write(urdf_content)
        else:
            urdf_file.write(no_mesh_urdf_content)

    print(f'URDF file for {mesh_name} created at {urdf_file_path}')



def create_urdf_gltf(mesh_name, save_dir = ASSET_STORE_DIR, sdf = False, scale = 1.0):
    # Create the URDF file path
    urdf_file_path = os.path.join(save_dir, f'{mesh_name}.urdf')

    # Get the mesh file path
    mesh_file_path = os.path.join(save_dir, f'{mesh_name}.gltf')

    # Create the URDF content
    if sdf: 
        urdf_content = f"""<?xml version="1.0"?>
<robot name="{mesh_name}">
<link name="base_link">
    <visual>
    <geometry>
        <mesh filename="{mesh_file_path}" scale="{scale} {scale} {scale}"/>
    </geometry>
    </visual>
    <collision>
    <geometry>
        <mesh filename="{mesh_file_path}" scale="{scale} {scale} {scale}"/>
    </geometry>
    <sdf resolution="256"/>
    </collision>
</link>
</robot>
"""
        
    else:
        urdf_content = f"""<?xml version="1.0"?>
<robot name="{mesh_name}">
<link name="base_link">
    <visual>
    <geometry>
        <mesh filename="{mesh_file_path}" scale="{scale} {scale} {scale}"/>
    </geometry>
    </visual>
    <collision>
    <geometry>
        <mesh filename="{mesh_file_path}" scale="{scale} {scale} {scale}"/>
    </geometry>
    </collision>
</link>
</robot>
"""    
        
    # Write the URDF content to the file
    with open(urdf_file_path, 'w') as urdf_file:
        urdf_file.write(urdf_content)

    print(f'URDF file for {mesh_name} created at {urdf_file_path}')



def create_assembly_part_urdf(mesh_name, scale, save_dir):
    # Create the URDF file path
    urdf_file_path = os.path.join(save_dir, f'{mesh_name}.urdf')

    # Create the URDF content
    urdf_content = f"""<?xml version="1.0"?>
<robot name="{mesh_name}">
<link name="base_link">
    <visual>
    <geometry>
        <mesh filename="{mesh_name}_final.gltf" scale="{scale} {scale} {scale}"/>
    </geometry>
    </visual>
    <collision>
    <geometry>
        <mesh filename="{mesh_name}_final.gltf" scale="{scale} {scale} {scale}"/>
    </geometry>
    <sdf resolution="256"/>
    </collision>
</link>
</robot>
"""

    # Write the URDF content to the file
    with open(urdf_file_path, 'w') as urdf_file:
        urdf_file.write(urdf_content)

    print(f'URDF file for {mesh_name} created at {urdf_file_path}')

def check_unzip_installed():
    try:
        # Attempt to call the 'unzip' command
        result = subprocess.run(['unzip', '-v'], stdout=subprocess.PIPE, stderr=subprocess.PIPE)
        # Check if the command returned successfully
        if result.returncode == 0:
            return True
        else:
            return False
    except FileNotFoundError:
        return False


import grpc
import time
import numpy as np
from dart_physics.avp_stream3.grpc_msg import dexterity_pb2, dexterity_pb2_grpc

# Function to create a random 4x4 matrix
def create_random_matrix():
    return dexterity_pb2.Matrix4x4(
        m00=np.random.random(), m01=np.random.random(), m02=np.random.random(), m03=np.random.random(),
        m10=np.random.random(), m11=np.random.random(), m12=np.random.random(), m13=np.random.random(),
        m20=np.random.random(), m21=np.random.random(), m22=np.random.random(), m23=np.random.random(),
        m30=np.random.random(), m31=np.random.random(), m32=np.random.random(), m33=np.random.random(),
    )

# Function to create a random Skeleton with 24 joint matrices
def create_random_skeleton():
    skeleton = dexterity_pb2.Skeleton()
    for _ in range(24):
        skeleton.jointMatrices.append(create_random_matrix())
    return skeleton

# Function to create a random HandUpdate
def create_random_hand_update():
    left_hand = dexterity_pb2.Hand(
        wristMatrix=create_random_matrix(),
        skeleton=create_random_skeleton()
    )
    right_hand = dexterity_pb2.Hand(
        wristMatrix=create_random_matrix(),
        skeleton=create_random_skeleton()
    )
    head = create_random_matrix()

    return dexterity_pb2.HandUpdate(
        left_hand=left_hand,
        right_hand=right_hand,
        Head=head
    )

# Mock client class to stream hand updates to the server
class MockHandTrackingClient:
    def __init__(self, server_address='localhost:12350'):
        self.channel = grpc.insecure_channel(server_address)
        self.stub = dexterity_pb2_grpc.HandTrackingServiceStub(self.channel)

    def stream_hand_updates(self):
        try:
            # Stream random hand updates to the server
            for response in self.stub.StreamHandUpdatesandGetSimStates(self.generate_hand_updates()):
                print(f"Received sim states: {response}")
        except grpc.RpcError as e:
            print(f"gRPC error: {e}")

    def generate_hand_updates(self):
        while True:
            hand_update = create_random_hand_update()
            yield hand_update

    def check_health(self):
        try:
            response = self.stub.Check(dexterity_pb2.HealthCheckRequest())
            print(f"Health check status: {response.status}")
        except grpc.RpcError as e:
            print(f"gRPC error: {e}")

# Usage example
if __name__ == '__main__':
    client = MockHandTrackingClient()
    
    # Perform a health check
    client.check_health()
    
    # Start streaming hand updates
    client.stream_hand_updates()



import requests
from time import perf_counter

def test_generate_scene():
    url = "http://172.178.76.173:8080/generate-scene"
    data = {
        "text_prompt": "A forest scene."
    }
    try:
        start_time = perf_counter()
        response = requests.post(url, data=data)
        end_time = perf_counter()
        print(f"[INFO] POST /generate-scene completed in {end_time - start_time:.4f} seconds")

        if response.status_code == 200:
            print("[INFO] Scene generated successfully!")
            with open("output_scene.glb", "wb") as f:
                f.write(response.content)
            print("[INFO] GLB file saved as 'output_scene.glb'")
        else:
            print(f"[ERROR] Failed to generate scene. Status code: {response.status_code}")
            print(f"[ERROR] Error detail: {response.json()}")
    except Exception as e:
        print(f"[ERROR] An error occurred: {str(e)}")

if __name__ == "__main__":
    test_generate_scene()




import sys
from dart_physics.robot_cfgs import GeomInfo
from scipy.spatial.transform import Rotation as R
import random
from pathlib import Path
from dart_physics.utils.run_functions import * 

HERE = Path(__file__).parent.absolute()


def random_xyz(i, xmin, xmax, ymin, ymax, zmin = 0.1):
    return [random.uniform(xmin, xmax), random.uniform(ymin, ymax), zmin + 0.05 * i ]

def duplicate_objects(name, num_max): 

    objs =  [name + "_{}".format(i) if i>0 else name for i in range(num_max)]

    # replicate the object 
    import shutil 
    obj_path = f"{HERE}/../assets/custom/{name}/{name}"
    
    for i in range(len(objs)):
        copy_path = f"{HERE}/../assets/custom/{name}_{i}"
        try: 
            shutil.copytree(obj_path, copy_path)
            shutil.move(f"{copy_path}/{name}.xml", f"{copy_path}/{name}_{i}.xml")
            # change the model name in the xml file
        except: 
            pass 
        # change the xml file (name.xml) to (name{i}.xml)

    return objs

def reset_function(model, data, robot_cfg, task_cfg): 

    id = robot_cfg["obj_startidx"] 
    vid = robot_cfg["obj_startidx"] 

    for ii, obj in enumerate(task_cfg['objects']): 

        if obj in task_cfg["fixed_objects"]: 
            continue 

        new_pos = task_cfg["default_poses"][obj][:3]
        new_quat = R.from_euler('xyz', [0, 90, np.random.randint(180)], degrees = True).as_quat(scalar_first = True).tolist()

        print(obj, len(new_pos))
        data.qpos[id:id+3] = new_pos
        data.qpos[id+3:id+7] = new_quat
        data.qvel[vid:vid+6] = 0.0

        id += 7 
        vid += 6

inks = duplicate_objects("dell_ink_cartridge", 3) + duplicate_objects("canon_cartridge", 3)

task_cfg = { 
    "name": "cvs_wall", 

    "objects": ["cvs_wall_final2"] + inks,  

    "fixed_objects": ["cvs_wall_final2"],

    "cone": "pyramidal", 
    "o_solref": ".000000000000000001 1",
    "o_solimp": ".000001 1",
    "impratio": "100.0",
    "multiccd": "enable" ,
    "override": "enable",

    "default_poses": {
        "cvs_wall_final2":  [-0.0, 0.35, 0.15 ] + R.from_euler('xyz', [0, 0, 0], degrees = True).as_quat(scalar_first = True).tolist(), 
        # "canon_ink_cartridge": [0.1, 0.15, 0.1 ] + R.from_euler('xyz', [0, 90, 90], degrees = True).as_quat(scalar_first = True).tolist(),
        # "dell_ink_cartridge": [-0.1, 0.15, 0.1 ] + R.from_euler('xyz', [0, 90, 90], degrees = True).as_quat(scalar_first = True).tolist(),
        # "washable_crayola": [-0.1, 0.15, 0.3 ] + R.from_euler('xyz', [0, 0, 0], degrees = True).as_quat(scalar_first = True).tolist(), 
    }, 

    "reset_function": reset_function, 
    
}

task_cfg["default_poses"].update({obj: random_xyz(i, -0.2, 0.2, 0.1, 0.3) + R.from_euler('xyz', [0, 90, np.random.randint(360)], degrees = True).as_quat(scalar_first = True).tolist() for i, obj in enumerate(inks)})



if __name__ == "__main__":


    from runs import * 
    from utils.scene_gen import construct_scene
    import mujoco
    from loop_rate_limiters import RateLimiter


    robot = "dual_panda"
    robot_cfg = load_robot_cfg(robot)

    model = construct_scene(task_cfg, robot_cfg)
    data = mujoco.MjData(model)

    viewer = mujoco.viewer.launch_passive(
        model=model, data=data, show_left_ui=True, show_right_ui=True)
    
    mujoco.mj_resetDataKeyframe(model, data, 0)
    mujoco.mj_forward(model, data)
    mujoco.mj_camlight(model, data)

    rate = RateLimiter(500)
    while True:
        mujoco.mj_step(model, data)
        viewer.sync()
        rate.sleep()
    


import numpy as np 
from typing import * 
from dart_physics.avp_stream3.grpc_msg import *

def matrix4x4_to_proto(matrix: np.ndarray) -> dexterity_pb2.Matrix4x4:
    """
    input:  matrix: np.ndarray - a 4x4 numpy array
    output: handtracking_pb2.Matrix4x4 - a protobuf message
    does:   converts a 4x4 numpy array to a protobuf message


    message Matrix4x4 {
        float m00 = 1;
        float m01 = 2;
        float m02 = 3;
        float m03 = 4;
        float m10 = 5;
        float m11 = 6;
        float m12 = 7;
        float m13 = 8;
        float m20 = 9;
        float m21 = 10;
        float m22 = 11;
        float m23 = 12;
        float m30 = 13;
        float m31 = 14;
        float m32 = 15;
        float m33 = 16;
        }

    """
    proto_matrix = dexterity_pb2.Matrix4x4()
    proto_matrix.m00 = matrix[0, 0]
    proto_matrix.m01 = matrix[0, 1]
    proto_matrix.m02 = matrix[0, 2]
    proto_matrix.m03 = matrix[0, 3]
    proto_matrix.m10 = matrix[1, 0]
    proto_matrix.m11 = matrix[1, 1]
    proto_matrix.m12 = matrix[1, 2]
    proto_matrix.m13 = matrix[1, 3]
    proto_matrix.m20 = matrix[2, 0]
    proto_matrix.m21 = matrix[2, 1]
    proto_matrix.m22 = matrix[2, 2]
    proto_matrix.m23 = matrix[2, 3] 
    # proto_matrix.m30 = matrix[3, 0]
    # proto_matrix.m31 = matrix[3, 1]
    # proto_matrix.m32 = matrix[3, 2]
    # proto_matrix.m33 = matrix[3, 3]

    return proto_matrix



def process_matrix(message):
    m = np.array([[[message.m00, message.m01, message.m02, message.m03],
                    [message.m10, message.m11, message.m12, message.m13],
                    [message.m20, message.m21, message.m22, message.m23],
                    [0, 0, 0, 1]]])
    return m 

def process_matrices(skeleton, matrix = np.eye(4)):
    return np.concatenate([matrix @ process_matrix(joint) for joint in skeleton], axis = 0)


def rotate_head(R, degrees=-90):
    # Convert degrees to radians
    theta = np.radians(degrees)
    # Create the rotation matrix for rotating around the x-axis
    R_x = np.array([[
        [1, 0, 0, 0],
        [0, np.cos(theta), -np.sin(theta), 0],
        [0, np.sin(theta), np.cos(theta), 0],
        [0, 0, 0, 1]
    ]])
    R_rotated = R @ R_x 
    return R_rotated


def get_pinch_distance(finger_messages): 
    fingers = process_matrices(finger_messages)
    thumb = fingers[4, :3, 3]
    index = fingers[9, :3, 3]

    return np.linalg.norm(thumb - index)

def get_wrist_roll(mat):
    """
    returns roll, pitch, yaw in radians
    """
    R = mat[0, :3, :3]

    # Calculate angles for rotation around z and y axis to align the first column with [1, 0, 0]
    # Angle to rotate around z-axis to align the projection on the XY plane
    theta_z = np.arctan2(R[1, 0], R[0, 0])  # arctan2(y, x)

    # Rotate R around the z-axis by -theta_z to align its x-axis on the XY plane
    Rz = np.array([
        [np.cos(-theta_z), -np.sin(-theta_z), 0],
        [np.sin(-theta_z), np.cos(-theta_z), 0],
        [0, 0, 1]
    ])
    R_after_z = Rz @ R

    # Angle to rotate around y-axis to align the x-axis with the global x-axis
    theta_y = np.arctan2(R_after_z[0, 2], R_after_z[0, 0])  # arctan2(z, x)

    # Since the goal is to align the x-axis, the rotation around the x-axis might not be necessary
    # unless there are specific orientations required for the y and z axes after the alignment.

    # Calculated angles (converted to degrees for easier interpretation)
    theta_z_deg = np.degrees(theta_z)
    theta_y_deg = np.degrees(theta_y)

    Ry = np.array([
        [np.cos(-theta_y), 0, np.sin(-theta_y)],
        [0, 1, 0],
        [-np.sin(-theta_y), 0, np.cos(-theta_y)]
    ])
    R_after_y = Ry @ R_after_z

    # Angle to rotate around x-axis to align the y-axis and z-axis properly with the global y-axis and z-axis
    theta_x = np.arctan2(R_after_y[1, 2], R_after_y[1, 1])  # arctan2(z, y) of the second row

    # Calculated angle (converted to degrees for easier interpretation)
    # theta_x_deg = np.degrees(theta_x)

    return theta_x 



"""Posture task implementation."""

from __future__ import annotations

from typing import Optional

import mujoco
import numpy as np
import numpy.typing as npt

from ..configuration import Configuration
from ..utils import get_freejoint_dims
from .exceptions import InvalidTarget, TargetNotSet, TaskDefinitionError
from .task import Task


class PostureTask(Task):
    """Regulate the joint angles of the robot towards a desired posture.

    A posture is a vector of actuated joint angles. Floating-base coordinates are not
    affected by this task.

    Attributes:
        target_q: Target configuration.
    """

    target_q: Optional[np.ndarray]

    def __init__(
        self,
        model: mujoco.MjModel,
        cost: npt.ArrayLike,
        gain: float = 1.0,
        lm_damping: float = 0.0,
    ):
        super().__init__(
            cost=np.zeros((model.nv,)),
            gain=gain,
            lm_damping=lm_damping,
        )
        self.target_q = None

        self._v_ids: np.ndarray | None
        _, v_ids_or_none = get_freejoint_dims(model)
        if v_ids_or_none:
            self._v_ids = np.asarray(v_ids_or_none)
        else:
            self._v_ids = None

        self.k = model.nv
        self.nq = model.nq
        self.set_cost(cost)

    def disable_joints(self, joint_ids: list[int]) -> None:
        # add joint ids to self._v_ids
        if self._v_ids is None:
            self._v_ids = np.array(joint_ids)
        else:
            self._v_ids = np.concatenate([self._v_ids, joint_ids])

    def set_cost(self, cost: npt.ArrayLike) -> None:
        cost = np.atleast_1d(cost)
        if cost.ndim != 1 or cost.shape[0] not in (1, self.k):
            raise TaskDefinitionError(
                f"{self.__class__.__name__} cost must be a vector of shape (1,) "
                f"(aka identical cost for all dofs) or ({self.k},). Got {cost.shape}"
            )
        if not np.all(cost >= 0.0):
            raise TaskDefinitionError(f"{self.__class__.__name__} cost should be >= 0")
        self.cost[: self.k] = cost

    def set_target(self, target_q: npt.ArrayLike) -> None:
        """Set the target posture.

        Args:
            target_q: Desired joint configuration.
        """
        target_q = np.atleast_1d(target_q)
        if target_q.ndim != 1 or target_q.shape[0] != (self.nq):
            raise InvalidTarget(
                f"Expected target posture to have shape ({self.nq},) but got "
                f"{target_q.shape}"
            )
        self.target_q = target_q.copy()

    def set_target_from_configuration(self, configuration: Configuration) -> None:
        """Set the target posture from the current configuration.

        Args:
            configuration: Robot configuration :math:`q`.
        """
        self.set_target(configuration.q)

    def compute_error(self, configuration: Configuration) -> np.ndarray:
        r"""Compute the posture task error.

        The error is defined as:

        .. math::

            e(q) = q^* \ominus q

        Args:
            configuration: Robot configuration :math:`q`.

        Returns:
            Posture task error vector :math:`e(q)`.
        """
        if self.target_q is None:
            raise TargetNotSet(self.__class__.__name__)

        # NOTE: mj_differentiatePos calculates qpos2 ⊖ qpos1.
        qvel = np.empty(configuration.nv)
        mujoco.mj_differentiatePos(
            m=configuration.model,
            qvel=qvel,
            dt=1.0,
            qpos1=configuration.q,
            qpos2=self.target_q,
        )

        if self._v_ids is not None:
            qvel[self._v_ids] = 0.0

        return qvel

    def compute_jacobian(self, configuration: Configuration) -> np.ndarray:
        r"""Compute the posture task Jacobian.

        The task Jacobian is defined as:

        .. math::

            J(q) = I_{n_v}

        Args:
            configuration: Robot configuration :math:`q`.

        Returns:
            Posture task jacobian :math:`J(q)`.
        """
        if self.target_q is None:
            raise TargetNotSet(self.__class__.__name__)

        jac = -np.eye(configuration.nv)
        if self._v_ids is not None:
            jac[:, self._v_ids] = 0.0

        return jac




import numpy as np 
from dart_physics.utils.run_functions import * 
from pathlib import Path
from dart_physics.robot_cfgs import * 

_HERE = Path(__file__).resolve().parent


allegro_cfg = {

    "robot": { 
        "l_robot": RobotConfig( 
                    xml_path = (_HERE / "../assets" / "wonik_allegro" / "left_hand.xml").as_posix(),
                    home_q = np.array([-0.5, 0.0, 0.3, 0, 0, 0, 1] + [0.] * 16), 
                    freejoint = True, 
                ), 
        "r_robot": RobotConfig(
                    xml_path = (_HERE / "../assets" / "wonik_allegro" / "right_hand.xml").as_posix(),
                    home_q = np.array([0.5, 0.0, 0.3, 0, 0, 0, 1] + [0.] * 16),
                    freejoint = True,
                ) 
    }, 
        
    "points" : { 
        "left_wrist":  FingerKeypoints("left", 0, "left_wrist_target", "l_robot/palm", weld = True, avp_transform = rot_z(90) @ rot_x(90)), 
        "right_wrist": FingerKeypoints("right", 0, "right_wrist_target", "r_robot/palm", weld = True), 

        "left_ff":     FingerKeypoints("left", 9, "left_ff_tip_target", "l_robot/ff_tip", type = "site"),
        "left_mf":      FingerKeypoints("left", 14, "left_mf_tip_target", "l_robot/mf_tip", type = "site"),
        "left_rf":      FingerKeypoints("left", 19, "left_rf_tip_target", "l_robot/rf_tip", type = "site"),
        "left_th":      FingerKeypoints("left", 4, "left_th_tip_target", "l_robot/th_tip", type = "site"),
        
        "right_ff":     FingerKeypoints("right", 9, "right_ff_tip_target", "r_robot/ff_tip", type = "site"),
        "right_mf":     FingerKeypoints("right", 14, "right_mf_tip_target", "r_robot/mf_tip", type = "site"),
        "right_rf":     FingerKeypoints("right", 19, "right_rf_tip_target", "r_robot/rf_tip", type = "site"),
        "right_th":     FingerKeypoints("right", 4, "right_th_tip_target", "r_robot/th_tip", type = "site"),
    },

    "avp_calib": { 

        "left_ff": {"scale": 1.4, "offset": np.array([-0.02, 0.0, 0.0])},
        "left_mf": {"scale": 1.3, "offset": np.array([+0.00, 0.0, 0.0])},
        "left_rf": {"scale": 1.3, "offset": np.array([+0.02, 0.0, 0.0])},
        "left_th": {"scale": 1.0, "offset": np.array([+0.00, 0.0, 0.0])},

        "right_ff": {"scale": 1.2, "offset": np.array([0.0, -0.02, 0.0])},
        "right_mf": {"scale": 1.2, "offset": np.array([0.0, -0.02, 0.0])},
        "right_rf": {"scale": 1.3, "offset": np.array([0.0, +0.02, 0.0])},
        "right_th": {"scale": 1.0, "offset": np.array([0.0, +0.00, 0.0])},

    }, 

    "ik_task" : [
        IKTasks(root = "left_wrist", target = "left_rf", scale = 1.3), 
        IKTasks(root = "left_wrist", target = "left_mf", scale = 1.2),
        IKTasks(root = "left_wrist", target = "left_ff", scale = 1.2),
        IKTasks(root = "left_wrist", target = "left_th", scale = 1.0),

        IKTasks(root = "right_wrist", target = "right_rf", scale = 1.3),
        IKTasks(root = "right_wrist", target = "right_mf", scale = 1.2),
        IKTasks(root = "right_wrist", target = "right_ff", scale = 1.2),
        IKTasks(root = "right_wrist", target = "right_th", scale = 1.0),
    ],

    "joints" : [ 
        # left hand 
        "l_robot/rfj0", "l_robot/rfj1", "l_robot/rfj2", "l_robot/rfj3",
        "l_robot/mfj0", "l_robot/mfj1", "l_robot/mfj2", "l_robot/mfj3",
        "l_robot/ffj0", "l_robot/ffj1", "l_robot/ffj2", "l_robot/ffj3",
        "l_robot/thj0", "l_robot/thj1", "l_robot/thj2", "l_robot/thj3",

        # right hand
        "r_robot/rfj0", "r_robot/rfj1", "r_robot/rfj2", "r_robot/rfj3",
        "r_robot/mfj0", "r_robot/mfj1", "r_robot/mfj2", "r_robot/mfj3",
        "r_robot/ffj0", "r_robot/ffj1", "r_robot/ffj2", "r_robot/ffj3",
        "r_robot/thj0", "r_robot/thj1", "r_robot/thj2", "r_robot/thj3",
    ],

    "bodies":  {
        # ff 
        "l_robot/ff_base":   "left_allegro_ff_base",   "l_robot/ff_distal":   "left_allegro_ff_distal", 
        "l_robot/ff_medial": "left_allegro_ff_medial", "l_robot/ff_proximal": "left_allegro_ff_proximal", 
        "l_robot/ff_tip": "left_allegro_ff_tip", 

        # mf 
        "l_robot/mf_base":   "left_allegro_mf_base",   "l_robot/mf_distal":   "left_allegro_mf_distal", 
        "l_robot/mf_medial": "left_allegro_mf_medial", "l_robot/mf_proximal": "left_allegro_mf_proximal", 
        "l_robot/mf_tip": "left_allegro_mf_tip", 

        # rf 
        "l_robot/rf_base":   "left_allegro_rf_base",   "l_robot/rf_distal":   "left_allegro_rf_distal",
        "l_robot/rf_medial": "left_allegro_rf_medial", "l_robot/rf_proximal": "left_allegro_rf_proximal",
        "l_robot/rf_tip": "left_allegro_rf_tip",

        # th
        "l_robot/th_base":   "left_allegro_th_base",   "l_robot/th_distal":   "left_allegro_th_distal",
        "l_robot/th_medial": "left_allegro_th_medial", "l_robot/th_proximal": "left_allegro_th_proximal",
        "l_robot/th_tip": "left_allegro_th_tip",

    } 

} 

mocap2frame = { 
    "dual_free_shadow": { 
        "lh_thdistal_target": "l_robot/lh_thdistal",
        "lh_ffdistal_target": "l_robot/lh_ffdistal",
        "lh_mfdistal_target": "l_robot/lh_mfdistal",
        "lh_rfdistal_target": "l_robot/lh_rfdistal",
        "lh_lfdistal_target": "l_robot/lh_lfdistal",

        "rh_thdistal_target": "r_robot/rh_thdistal",
        "rh_ffdistal_target": "r_robot/rh_ffdistal",
        "rh_mfdistal_target": "r_robot/rh_mfdistal",
        "rh_rfdistal_target": "r_robot/rh_rfdistal",
        "rh_lfdistal_target": "r_robot/rh_lfdistal",
    }, 

    "dual_free_allegro":  { 
        "left_rf_tip_target": "l_robot/rf_tip",
        "left_mf_tip_target": "l_robot/mf_tip",
        "left_ff_tip_target": "l_robot/ff_tip",
        "left_th_tip_target": "l_robot/th_tip",

        "right_rf_tip_target": "r_robot/rf_tip",
        "right_mf_tip_target": "r_robot/mf_tip",
        "right_ff_tip_target": "r_robot/ff_tip",
        "right_th_tip_targe,t": "r_robot/th_tip",
    }
    
}

weld = {
    "dual_free_shadow": { 
        "left_wrist_target": "l_robot/lh_wrist",
        "right_wrist_target": "r_robot/rh_wrist",
    }, 

    "dual_free_allegro":  { 
        "left_wrist_target": "l_robot/palm", 
        "right_wrist_target": "r_robot/palm"

    }
}

ikrel = {
    # root: frame 
    "dual_free_allegro": { 
        "l_robot/palm": "l_robot/rf_tip", 
        "l_robot/palm": "l_robot/"
    }
}


avp2mocap = { 
    "dual_free_shadow": { 
        "right": {
            0 : "right_wrist_target", 
            4 : "rh_thdistal_target",
            9 : "rh_ffdistal_target",
            14: "rh_mfdistal_target",
            19: "rh_rfdistal_target",
            24: "rh_lfdistal_target",
        }, 
        "left": {
            0 : "left_wrist_target", 
            4 : "lh_thdistal_target",
            9 : "lh_ffdistal_target",
            14: "lh_mfdistal_target",
            19: "lh_rfdistal_target",
            24: "lh_lfdistal_target",
        },
    }, 

    "dual_free_allegro": { 
        "left": {
            "0": "ri"
        }
    }
}


joints = { 
    "dual_free_shadow": [ 
                            "r_robot/rh_FFJ4",
                          "r_robot/rh_FFJ3",
                            "r_robot/rh_FFJ2",
                            "r_robot/rh_FFJ1",
                            "r_robot/rh_MFJ4",
                            "r_robot/rh_MFJ3",
                            "r_robot/rh_MFJ2",
                            "r_robot/rh_MFJ1",
                            "r_robot/rh_RFJ4",
                            "r_robot/rh_RFJ3",
                            "r_robot/rh_RFJ2",
                            "r_robot/rh_RFJ1",
                            "r_robot/rh_LFJ5",
                            "r_robot/rh_LFJ4",
                            "r_robot/rh_LFJ3",
                            "r_robot/rh_LFJ2",
                            "r_robot/rh_LFJ1",
                            "r_robot/rh_THJ5",
                            "r_robot/rh_THJ4",
                            "r_robot/rh_THJ3",
                            "r_robot/rh_THJ2",
                            "r_robot/rh_THJ1",

                            "l_robot/lh_FFJ4",
                            "l_robot/lh_FFJ3",
                            "l_robot/lh_FFJ2",
                            "l_robot/lh_FFJ1",
                            "l_robot/lh_MFJ4",
                            "l_robot/lh_MFJ3",
                            "l_robot/lh_MFJ2",
                            "l_robot/lh_MFJ1",
                            "l_robot/lh_RFJ4",
                            "l_robot/lh_RFJ3",
                            "l_robot/lh_RFJ2",
                            "l_robot/lh_RFJ1",
                            "l_robot/lh_LFJ5",
                            "l_robot/lh_LFJ4",
                            "l_robot/lh_LFJ3",
                            "l_robot/lh_LFJ2",
                            "l_robot/lh_LFJ1",
                            "l_robot/lh_THJ5",
                            "l_robot/lh_THJ4",
                            "l_robot/lh_THJ3",
                            "l_robot/lh_THJ2",
                            "l_robot/lh_THJ1"
                    ]
}


"""Kinematic limits."""

from .collision_avoidance_limit import CollisionAvoidanceLimit
from .configuration_limit import ConfigurationLimit
from .exceptions import LimitDefinitionError
from .limit import Constraint, Limit
from .velocity_limit import VelocityLimit

__all__ = (
    "ConfigurationLimit",
    "CollisionAvoidanceLimit",
    "Constraint",
    "Limit",
    "VelocityLimit",
    "LimitDefinitionError",
)



from __future__ import annotations

from dataclasses import dataclass

import mujoco
import numpy as np

from ..exceptions import InvalidMocapBody
from .base import MatrixLieGroup
from .so3 import SO3
from .utils import get_epsilon, skew

_IDENTITY_WXYZ_XYZ = np.array([1.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0], dtype=np.float64)


@dataclass(frozen=True)
class SE3(MatrixLieGroup):
    """Special Euclidean group for proper rigid transforms in 3D.

    Internal parameterization is (qw, qx, qy, qz, x, y, z). Tangent parameterization is
    (vx, vy, vz, omega_x, omega_y, omega_z).
    """

    wxyz_xyz: np.ndarray
    matrix_dim: int = 4
    parameters_dim: int = 7
    tangent_dim: int = 6
    space_dim: int = 3

    def __repr__(self) -> str:
        quat = np.round(self.wxyz_xyz[:4], 5)
        xyz = np.round(self.wxyz_xyz[4:], 5)
        return f"{self.__class__.__name__}(wxyz={quat}, xyz={xyz})"

    def copy(self) -> SE3:
        return SE3(wxyz_xyz=np.array(self.wxyz_xyz))

    def parameters(self) -> np.ndarray:
        return self.wxyz_xyz

    @classmethod
    def identity(cls) -> SE3:
        return SE3(wxyz_xyz=_IDENTITY_WXYZ_XYZ)

    @classmethod
    def from_rotation_and_translation(
        cls,
        rotation: SO3,
        translation: np.ndarray,
    ) -> SE3:
        assert translation.shape == (SE3.space_dim,)
        return SE3(wxyz_xyz=np.concatenate([rotation.wxyz, translation]))

    @classmethod
    def from_rotation(cls, rotation: SO3) -> SE3:
        return SE3.from_rotation_and_translation(
            rotation=rotation,
            translation=np.zeros(
                SE3.space_dim,
            ),
        )

    @classmethod
    def from_translation(cls, translation: np.ndarray) -> SE3:
        return SE3.from_rotation_and_translation(
            rotation=SO3.identity(), translation=translation
        )

    @classmethod
    def from_matrix(cls, matrix: np.ndarray) -> SE3:
        assert matrix.shape == (SE3.matrix_dim, SE3.matrix_dim)
        return SE3.from_rotation_and_translation(
            rotation=SO3.from_matrix(matrix[:3, :3]),
            translation=matrix[:3, 3],
        )

    @classmethod
    def from_mocap_id(cls, data: mujoco.MjData, mocap_id: int) -> SE3:
        return SE3.from_rotation_and_translation(
            rotation=SO3(data.mocap_quat[mocap_id]),
            translation=data.mocap_pos[mocap_id],
        )

    @classmethod
    def from_mocap_name(
        cls, model: mujoco.MjModel, data: mujoco.MjData, mocap_name: str
    ) -> SE3:
        mocap_id = model.body(mocap_name).mocapid[0]
        if mocap_id == -1:
            raise InvalidMocapBody(mocap_name, model)
        return SE3.from_mocap_id(data, mocap_id)

    @classmethod
    def sample_uniform(cls) -> SE3:
        return SE3.from_rotation_and_translation(
            rotation=SO3.sample_uniform(),
            translation=np.random.uniform(-1.0, 1.0, size=(SE3.space_dim,)),
        )

    def rotation(self) -> SO3:
        return SO3(wxyz=self.wxyz_xyz[:4])

    def translation(self) -> np.ndarray:
        return self.wxyz_xyz[4:]

    def as_matrix(self) -> np.ndarray:
        hmat = np.eye(self.matrix_dim, dtype=np.float64)
        hmat[:3, :3] = self.rotation().as_matrix()
        hmat[:3, 3] = self.translation()
        return hmat

    @classmethod
    def exp(cls, tangent: np.ndarray) -> SE3:
        assert tangent.shape == (SE3.tangent_dim,)
        rotation = SO3.exp(tangent[3:])
        theta_squared = tangent[3:] @ tangent[3:]
        use_taylor = theta_squared < get_epsilon(theta_squared.dtype)
        theta_squared_safe = 1.0 if use_taylor else theta_squared
        theta_safe = np.sqrt(theta_squared_safe)
        skew_omega = skew(tangent[3:])
        if use_taylor:
            V = rotation.as_matrix()
        else:
            V = (
                np.eye(3, dtype=np.float64)
                + (1.0 - np.cos(theta_safe)) / (theta_squared_safe) * skew_omega
                + (theta_safe - np.sin(theta_safe))
                / (theta_squared_safe * theta_safe)
                * (skew_omega @ skew_omega)
            )
        return SE3.from_rotation_and_translation(
            rotation=rotation,
            translation=V @ tangent[:3],
        )

    def inverse(self) -> SE3:
        R_inv = self.rotation().inverse()
        return SE3.from_rotation_and_translation(
            rotation=R_inv,
            translation=-(R_inv @ self.translation()),
        )

    def normalize(self) -> SE3:
        return SE3.from_rotation_and_translation(
            rotation=self.rotation().normalize(),
            translation=self.translation(),
        )

    def apply(self, target: np.ndarray) -> np.ndarray:
        assert target.shape == (SE3.space_dim,)
        return self.rotation() @ target + self.translation()

    def multiply(self, other: SE3) -> SE3:
        return SE3.from_rotation_and_translation(
            rotation=self.rotation() @ other.rotation(),
            translation=(self.rotation() @ other.translation()) + self.translation(),
        )

    def log(self) -> np.ndarray:
        omega = self.rotation().log()
        theta_squared = omega @ omega
        use_taylor = theta_squared < get_epsilon(theta_squared.dtype)
        skew_omega = skew(omega)
        theta_squared_safe = 1.0 if use_taylor else theta_squared
        theta_safe = np.sqrt(theta_squared_safe)
        half_theta_safe = 0.5 * theta_safe
        skew_omega_norm = skew_omega @ skew_omega
        if use_taylor:
            V_inv = (
                np.eye(3, dtype=np.float64) - 0.5 * skew_omega + skew_omega_norm / 12.0
            )
        else:
            V_inv = (
                np.eye(3, dtype=np.float64)
                - 0.5 * skew_omega
                + (
                    1.0
                    - theta_safe
                    * np.cos(half_theta_safe)
                    / (2.0 * np.sin(half_theta_safe))
                )
                / theta_squared_safe
                * skew_omega_norm
            )
        return np.concatenate([V_inv @ self.translation(), omega])

    def adjoint(self) -> np.ndarray:
        R = self.rotation().as_matrix()
        return np.block(
            [
                [R, skew(self.translation()) @ R],
                [np.zeros((3, 3), dtype=np.float64), R],
            ]
        )

    # Jacobians.

    # Eqn 179 a)
    @classmethod
    def ljac(cls, other: np.ndarray) -> np.ndarray:
        theta = other[3:]
        if theta @ theta < get_epsilon(theta.dtype):
            return np.eye(cls.tangent_dim)
        Q = _getQ(other)
        J = SO3.ljac(theta)
        O = np.zeros((3, 3))
        return np.block([[J, Q], [O, J]])

    # Eqn 179 b)
    @classmethod
    def ljacinv(cls, other: np.ndarray) -> np.ndarray:
        theta = other[3:]
        if theta @ theta < get_epsilon(theta.dtype):
            return np.eye(cls.tangent_dim)
        Q = _getQ(other)
        J_inv = SO3.ljacinv(theta)
        O = np.zeros((3, 3))
        return np.block([[J_inv, -J_inv @ Q @ J_inv], [O, J_inv]])


# Eqn 180.
def _getQ(c) -> np.ndarray:
    theta_sq = c[3:] @ c[3:]
    A = 0.5
    if theta_sq < get_epsilon(theta_sq.dtype):
        B = (1.0 / 6.0) + (1.0 / 120.0) * theta_sq
        C = -(1.0 / 24.0) + (1.0 / 720.0) * theta_sq
        D = -(1.0 / 60.0)
    else:
        theta = np.sqrt(theta_sq)
        sin_theta = np.sin(theta)
        cos_theta = np.cos(theta)
        B = (theta - sin_theta) / (theta_sq * theta)
        C = (1.0 - theta_sq / 2.0 - cos_theta) / (theta_sq * theta_sq)
        D = ((2) * theta - (3) * sin_theta + theta * cos_theta) / (
            (2) * theta_sq * theta_sq * theta
        )
    V = skew(c[:3])
    W = skew(c[3:])
    VW = V @ W
    WV = VW.T
    WVW = WV @ W
    VWW = VW @ W
    return (
        +A * V
        + B * (WV + VW + WVW)
        - C * (VWW - VWW.T - 3 * WVW)
        + D * (WVW @ W + W @ WVW)
    )



"""All kinematic limits derive from the :class:`Limit` base class."""

import abc
from typing import NamedTuple, Optional

import numpy as np

from ..configuration import Configuration


class Constraint(NamedTuple):
    r"""Linear inequality constraint of the form :math:`G(q) \Delta q \leq h(q)`.

    Inactive if G and h are None.
    """

    G: Optional[np.ndarray] = None
    """Shape (nv, nv)."""
    h: Optional[np.ndarray] = None
    """Shape (nv,)."""

    @property
    def inactive(self) -> bool:
        """Returns True if the constraint is inactive."""
        return self.G is None and self.h is None


class Limit(abc.ABC):
    """Abstract base class for kinematic limits.

    Subclasses must implement the :py:meth:`~Limit.compute_qp_inequalities` method
    which takes in the current robot configuration and integration time step and
    returns an instance of :class:`Constraint`.
    """

    @abc.abstractmethod
    def compute_qp_inequalities(
        self,
        configuration: Configuration,
        dt: float,
    ) -> Constraint:
        r"""Compute limit as linearized QP inequalities of the form:

        .. math::

            G(q) \Delta q \leq h(q)

        where :math:`q \in {\cal C}` is the robot's configuration and
        :math:`\Delta q \in T_q({\cal C})` is the displacement in the tangent
        space at :math:`q`.

        Args:
            configuration: Robot configuration :math:`q`.
            dt: Integration time step in [s].

        Returns:
            Pair :math:`(G, h)`.
        """
        raise NotImplementedError



import zmq
import threading
import time 

class ALOHAZMQSubscriber(threading.Thread):
    def __init__(self, zmq_port):
        """
        Initializes the ZMQSubscriber, connects to the ZMQ publisher, and starts the thread.
        
        :param zmq_port: The ZMQ port to connect to for receiving data.
        """
        super(ALOHAZMQSubscriber, self).__init__()
        self.zmq_port = zmq_port
        self.latest = None  # Store the latest received data
        self.keep_running = True
        
        # Setup ZMQ subscriber socket
        context = zmq.Context()
        self.socket = context.socket(zmq.SUB)
        self.socket.connect(f"tcp://128.31.35.92:{zmq_port}")
        self.socket.setsockopt_string(zmq.SUBSCRIBE, '')  # Subscribe to all messages

    def run(self):
        """
        Run method to receive ZMQ messages in the background and continuously update `self.latest`.
        """
        while self.keep_running:
            try:
                # Receive the message (deserialized NumPy array)
                self.latest = self.socket.recv_pyobj()

            except zmq.ZMQError as e:
                print(f"ZMQ Error: {e}")
                break
    
    def stop(self):
        """
        Stop the ZMQ subscriber thread.
        """
        self.keep_running = False
        self.socket.close()

# Usage example in your main launch script
if __name__ == "__main__":
    zmq_port = 5555  # Replace with the port you're using for ZMQ

    # Instantiate the subscriber
    zmq_sub = ALOHAZMQSubscriber(zmq_port)

    # Start the subscriber thread
    zmq_sub.start()

    try:
        while True:
            # Read the latest data at your own pace
            if zmq_sub.latest is not None:
                print(f"Latest concatenated joint data: {zmq_sub.latest}")
            
            # Simulate doing other tasks in your main thread
            time.sleep(1)

    except KeyboardInterrupt:
        print("Shutting down...")
    
    finally:
        # Stop the ZMQ subscriber thread
        zmq_sub.stop()
        zmq_sub.join()



"""Configuration space of a robot model.

The :class:`Configuration` class encapsulates a MuJoCo
`model <https://mujoco.readthedocs.io/en/latest/APIreference/APItypes.html#mjmodel>`__
and `data <https://mujoco.readthedocs.io/en/latest/APIreference/APItypes.html#mjdata>`__,
offering easy access to frame transforms and frame Jacobians. A frame refers to a coordinate
system that can be attached to various parts of the robot, such as a body, geom, or site.
"""

from typing import Optional

import mujoco
import numpy as np

from . import constants as consts
from . import exceptions
from .lie import SE3, SO3


class Configuration:
    """Encapsulates a model and data for convenient access to kinematic quantities.

    This class provides methods to access and update the kinematic quantities of a robot
    model, such as frame transforms and Jacobians. It performs forward kinematics at every
    time step, ensuring up-to-date information about the robot's state.

    Key functionalities include:

    * Running forward kinematics to update the state.
    * Checking configuration limits.
    * Computing Jacobians for different frames.
    * Retrieving frame transforms relative to the world frame.
    * Integrating velocities to update configurations.
    """

    def __init__(
        self,
        model: mujoco.MjModel,
        q: Optional[np.ndarray] = None,
    ):
        """Constructor.

        Args:
            model: Mujoco model.
            q: Configuration to initialize from. If None, the configuration
                is initialized to the default configuration `qpos0`.
        """
        self.model = model
        self.data = mujoco.MjData(model)
        self.update(q=q)

    def update(self, q: Optional[np.ndarray] = None) -> None:
        """Run forward kinematics.

        Args:
            q: Optional configuration vector to override internal `data.qpos` with.
        """
        if q is not None:
            self.data.qpos = q
        # The minimal function call required to get updated frame transforms is
        # mj_kinematics. An extra call to mj_comPos is required for updated Jacobians.
        mujoco.mj_kinematics(self.model, self.data)
        mujoco.mj_comPos(self.model, self.data)

    def update_from_keyframe(self, key_name: str) -> None:
        """Update the configuration from a keyframe.

        Args:
            key_name: The name of the keyframe.
        """
        key_id = mujoco.mj_name2id(self.model, mujoco.mjtObj.mjOBJ_KEY, key_name)
        if key_id == -1:
            raise exceptions.InvalidKeyframe(key_name, self.model)
        self.update(q=self.model.key_qpos[key_id])

    def check_limits(self, tol: float = 1e-6, safety_break: bool = True) -> None:
        """Check that the current configuration is within bounds.

        Args:
            tol: Tolerance in [rad].
            safety_break: If True, stop execution and raise an exception if the current
                configuration is outside limits. If False, print a warning and continue
                execution.
        """
        for jnt in range(self.model.njnt):
            jnt_type = self.model.jnt_type[jnt]
            if (
                jnt_type == mujoco.mjtJoint.mjJNT_FREE
                or not self.model.jnt_limited[jnt]
            ):
                continue
            padr = self.model.jnt_qposadr[jnt]
            qval = self.q[padr]
            qmin = self.model.jnt_range[jnt, 0]
            qmax = self.model.jnt_range[jnt, 1]
            if qval < qmin - tol or qval > qmax + tol:
                if safety_break:
                    raise exceptions.NotWithinConfigurationLimits(
                        joint_id=jnt,
                        value=qval,
                        lower=qmin,
                        upper=qmax,
                        model=self.model,
                    )
                else:
                    # print(
                    #     f"Value {qval:.2f} at index {jnt} is outside of its limits: "
                    #     f"[{qmin:.2f}, {qmax:.2f}]"
                    # )
                    pass 

    def get_frame_jacobian(self, frame_name: str, frame_type: str) -> np.ndarray:
        r"""Compute the Jacobian matrix of a frame velocity.

        Denoting our frame by :math:`B` and the world frame by :math:`W`, the
        Jacobian matrix :math:`{}_B J_{WB}` is related to the body velocity
        :math:`{}_B v_{WB}` by:

        .. math::

            {}_B v_{WB} = {}_B J_{WB} \dot{q}

        Args:
            frame_name: Name of the frame in the MJCF.
            frame_type: Type of frame. Can be a geom, a body or a site.

        Returns:
            Jacobian :math:`{}_B J_{WB}` of the frame.
        """
        if frame_type not in consts.SUPPORTED_FRAMES:
            raise exceptions.UnsupportedFrame(frame_type, consts.SUPPORTED_FRAMES)

        frame_id = mujoco.mj_name2id(
            self.model, consts.FRAME_TO_ENUM[frame_type], frame_name
        )
        if frame_id == -1:
            raise exceptions.InvalidFrame(
                frame_name=frame_name,
                frame_type=frame_type,
                model=self.model,
            )

        jac = np.empty((6, self.model.nv))
        jac_func = consts.FRAME_TO_JAC_FUNC[frame_type]
        jac_func(self.model, self.data, jac[:3], jac[3:], frame_id)

        # MuJoCo jacobians have a frame of reference centered at the local frame but
        # aligned with the world frame. To obtain a jacobian expressed in the local
        # frame, aka body jacobian, we need to left-multiply by A[T_fw].
        xmat = getattr(self.data, consts.FRAME_TO_XMAT_ATTR[frame_type])[frame_id]
        R_wf = SO3.from_matrix(xmat.reshape(3, 3))
        A_fw = SE3.from_rotation(R_wf.inverse()).adjoint()
        jac = A_fw @ jac

        return jac

    def get_transform_frame_to_world(self, frame_name: str, frame_type: str) -> SE3:
        """Get the pose of a frame at the current configuration.

        Args:
            frame_name: Name of the frame in the MJCF.
            frame_type: Type of frame. Can be a geom, a body or a site.

        Returns:
            The pose of the frame in the world frame.
        """
        if frame_type not in consts.SUPPORTED_FRAMES:
            raise exceptions.UnsupportedFrame(frame_type, consts.SUPPORTED_FRAMES)

        frame_id = mujoco.mj_name2id(
            self.model, consts.FRAME_TO_ENUM[frame_type], frame_name
        )
        if frame_id == -1:
            raise exceptions.InvalidFrame(
                frame_name=frame_name,
                frame_type=frame_type,
                model=self.model,
            )

        xpos = getattr(self.data, consts.FRAME_TO_POS_ATTR[frame_type])[frame_id]
        xmat = getattr(self.data, consts.FRAME_TO_XMAT_ATTR[frame_type])[frame_id]
        return SE3.from_rotation_and_translation(
            rotation=SO3.from_matrix(xmat.reshape(3, 3)),
            translation=xpos,
        )

    def get_transform(
        self,
        source_name: str,
        source_type: str,
        dest_name: str,
        dest_type: str,
    ) -> SE3:
        """Get the pose of a frame with respect to another frame at the current
        configuration.

        Args:
            source_name: Name of the frame in the MJCF.
            source_type: Source type of frame. Can be a geom, a body or a site.
            dest_name: Name of the frame to get the pose in.
            dest_type: Dest type of frame. Can be a geom, a body or a site.

        Returns:
            The pose of `source_name` in `dest_name`.
        """
        transform_source_to_world = self.get_transform_frame_to_world(
            source_name, source_type
        )
        transform_dest_to_world = self.get_transform_frame_to_world(
            dest_name, dest_type
        )
        return transform_dest_to_world.inverse() @ transform_source_to_world

    def integrate(self, velocity: np.ndarray, dt: float) -> np.ndarray:
        """Integrate a velocity starting from the current configuration.

        Args:
            velocity: The velocity in tangent space.
            dt: Integration duration in [s].

        Returns:
            The new configuration after integration.
        """
        q = self.data.qpos.copy()
        mujoco.mj_integratePos(self.model, q, velocity, dt)
        return q

    def integrate_inplace(self, velocity: np.ndarray, dt: float) -> None:
        """Integrate a velocity and update the current configuration inplace.

        Args:
            velocity: The velocity in tangent space.
            dt: Integration duration in [s].
        """
        mujoco.mj_integratePos(self.model, self.data.qpos, velocity, dt)
        self.update()

    # Aliases.

    @property
    def q(self) -> np.ndarray:
        """The current configuration vector."""
        return self.data.qpos.copy()

    @property
    def nv(self) -> int:
        """The dimension of the tangent space."""
        return self.model.nv

    @property
    def nq(self) -> int:
        """The dimension of the configuration space."""
        return self.model.nq




import os 
import numpy as np 
import zmq 
from threading import Thread
from queue import Queue
import matplotlib.pyplot as plt
import time
from PIL import Image
import math
from scipy.spatial.transform import Rotation as R
from typing import * 

class time_profiler:

    def __init__(self): 
        """
        class that profiles the time taken by different functions 
        
        it'll be called as follows: 
        profiler = time_profiler()
        profiler.start("function_name")
        # do something
        profiler.end("function_name") <-- and this will print the time taken by the function 
        """
        self.timings = {}
        self.start_time = 0

    def start(self, name):
        self.start_time = time.time()
        self.timings[name] = 0

    def end(self, name):
        self.timings[name] = time.time() - self.start_time
        if os.getenv("MJ_DEBUG", False):
            print(f"{name} took {self.timings[name]}")

def pprint(msg, time): 

    if os.getenv("MJ_DEBUG", False):
        print(msg, time)

YUP2ZUP = np.array([[1, 0, 0, 0], 
                    [0, 0, -1, 0], 
                    [0, 1, 0, 0],
                    [0, 0, 0, 1]], dtype = np.float64)

def segmentation_mask_to_rgb(mask, color_map=None):
    """
    Convert a segmentation mask to an RGB image for debugging.
    
    Args:
        mask (numpy.ndarray): 2D array with segmentation mask values.
        color_map (dict): Optional. A dictionary mapping mask values to RGB tuples. 
                          If None, a default color map will be generated.
                          
    Returns:
        PIL.Image.Image: The RGB image.
    """
    # Generate a color map if none is provided
    if color_map is None:
        unique_values = np.unique(mask)
        np.random.seed(42)  # For reproducibility
        color_map = {val: tuple(np.random.randint(0, 256, 3).tolist()) for val in unique_values}
     
    # Create an RGB image using the color map
    rgb_image = np.zeros((mask.shape[0], mask.shape[1], 3), dtype=np.uint8)
    
    for val, color in color_map.items():
        rgb_image[mask == val] = color

    return Image.fromarray(rgb_image)


def data_receiver(socket, data_queue):
    while True:
        msg = socket.recv_pyobj()  # Receive the image data
        data_queue.put(msg)

def image_rendering_process(addr):
    context = zmq.Context()
    socket = context.socket(zmq.SUB)
    socket.connect(addr)
    socket.setsockopt(zmq.SUBSCRIBE, b'')

    data_queue = Queue()

    # Start the data receiving thread
    receiver_thread = Thread(target=data_receiver, args=(socket, data_queue))
    receiver_thread.daemon = True
    receiver_thread.start()

    plt.ion()  # Turn on interactive mode
    fig, ax = plt.subplots()

    # Initialize the image plot with some dummy data
    img = np.zeros((480, 1280, 3), dtype=np.uint8)  # Assuming a 480x640 RGB image
    im_plot = ax.imshow(img)

    plt.tight_layout()

    while True:
        while not data_queue.empty():
            img_data = data_queue.get()
            print("here", img_data.shape)

            # Update the image plot with the new image data
            im_plot.set_data(img_data)

            # If you want to adjust the color limits, you can do it like this:
            # im_plot.set_clim(vmin=-1, vmax=2)  # Adjust these limits based on your data

            ax.draw_artist(im_plot)
            fig.canvas.flush_events()  # Update the plot

        time.sleep(0.01)  # Small sleep to prevent overloading the CPU


def get_random_perturbed_pose(
    old_pose: np.ndarray,
    reset_spec: Dict[str, Any],
) -> np.ndarray:
    new_pose = old_pose.copy()
    
    # perturb position
    min_pos = np.array(reset_spec['min_pos'])
    max_pos = np.array(reset_spec['max_pos'])
    new_pose[:3] = old_pose[:3] + np.random.rand(3) * (max_pos - min_pos) + min_pos

    # get angle perturbation as euler angles
    max_euler = np.array(reset_spec['max_rot'])
    min_euler = np.array(reset_spec['min_rot'])
    euler_perturbation = np.random.rand(3) * (max_euler - min_euler) + min_euler

    # perturb orientation by euler angles
    r = R.from_euler('xyz', euler_perturbation, degrees=True)
    old_rot = R.from_quat(old_pose[3:])
    new_rot = r * old_rot
    new_pose[3:] = new_rot.as_quat()

    return new_pose



def convert_to_z_up(quaternion_str, table_height):
    # Parse the input string
    x, y, z, rw, rx, ry, rz = map(float, quaternion_str.split(','))

    mat = np.eye(4)
    mat[:3, :3] = R.from_quat([rw, rx, ry, rz], scalar_first=True).as_matrix()
    mat[:3, -1] = np.array([x, y, z])

    new_mat = YUP2ZUP @ mat 

    new_pos = new_mat[:3, -1]
    new_pos[2] -= table_height
    new_quat = R.from_matrix(new_mat[:3, :3]).as_quat(scalar_first=True)

    result = np.concatenate((new_pos, new_quat))

    return result

# Define rotation matrices for each axis
def rot_x(angle):
    angle = math.radians(angle)
    return np.array([[1, 0, 0, 0],
                     [0, math.cos(angle), -math.sin(angle), 0],
                     [0, math.sin(angle), math.cos(angle), 0],
                     [0, 0, 0, 1]])#.type(np.float64)

def rot_y(angle):
    angle = math.radians(angle)
    return np.array([[math.cos(angle), 0, math.sin(angle), 0],
                     [0, 1, 0, 0],
                     [-math.sin(angle), 0, math.cos(angle), 0],
                     [0, 0, 0, 1]])#.astype(np.float64)

def rot_z(angle):
    angle = math.radians(angle)
    return np.array([[math.cos(angle), -math.sin(angle), 0, 0],
                    [math.sin(angle), math.cos(angle), 0, 0],
                    [0, 0, 1, 0],
                    [0, 0, 0, 1]])#.astype(np.float64)

def sm2mat(sm_state, type = "right"): 
    # convert roll pitch yaw to rotation matrix

    if type == "right": 
        roll = - sm_state.roll * 0.1
        pitch = - sm_state.pitch * 0.1
        yaw = - sm_state.yaw * 0.1
        
        rot_mat = rot_z(yaw) @ rot_y(pitch) @ rot_x(roll)

    else:
        roll = - sm_state.roll * 0.1
        pitch = - sm_state.pitch * 0.1
        yaw = - sm_state.yaw * 0.1
        
        rot_mat = rot_z(yaw) @ rot_y(- roll) @ rot_x(pitch)

    mat = np.eye(4)
    mat[:3, :3] = rot_mat[:3, :3]
    # translations 
    mat[:3, -1] = np.array([sm_state.x * 0.001, sm_state.y* 0.001, sm_state.z* 0.001])

    return mat




import grpc
from concurrent import futures
import time
import threading
import numpy as np
from dart_physics.avp_stream3.grpc_msg import * 
from dart_physics.avp_stream3.utils.grpc_utils import * 
from flask import Flask, request, jsonify
from PIL import Image
import io 
import cv2 
from dart_physics.utils.run_functions import rot_x

YUP2ZUP = np.array([[[1, 0, 0, 0], 
                    [0, 0, -1, 0], 
                    [0, 1, 0, 0],
                    [0, 0, 0, 1]]], dtype = np.float64)

ZUP2YUP = np.array([[1, 0, 0, 0],
                    [0, 0, 1, 0],
                    [0, -1, 0, 0],
                    [0, 0, 0, 1]], dtype = np.float64)


class HandTrackingServer:

    def __init__(self, port=12350):
        self.port = port
        self.latest = None
        self.axis_transform = YUP2ZUP
        self.sim_states = dexterity_pb2.SimStates()
        self.sim_states.matrices["left_hand"].CopyFrom(matrix4x4_to_proto(np.eye(4)))
        
        self.convert = np.array([[-1. ,  0. ,  0. ,  0. ],
            [ 0. , -1. ,  0. ,  0.6],
            [ 0. ,  0. ,  1. ,  0. ],
            [ 0. ,  0. ,  0. ,  1. ]])
        
        self.thirdperson = False 

        # self.cap = cv2.VideoCapture(0) # MODIFY TO CORRECT PORT
        resolution = (1280, 480)
        # self.cap.set(cv2.CAP_PROP_FRAME_WIDTH, resolution[0])
        # self.cap.set(cv2.CAP_PROP_FRAME_HEIGHT, resolution[1])

        ret = False
        # while not ret: 

        #     ret, image = self.cap.read() # reads frame that has stereo images side by side
        #     if not ret:
        #         print("failed to grab")
        #         return 
            
        #     image = cv2.rotate(image, cv2.ROTATE_180)
        #     # print("real_image:", image.shape)
        #     w = image.shape[1]//2
        #     h = image.shape[0]
        #     im_left = image[:, :w]
        #     im_right = image[:, w:2*w]
            
        #     _, buffer = cv2.imencode('.jpg', im_left)
        #     _, buffer2 = cv2.imencode('.jpg', im_right)
        #     # print("encoded both")
            
        #     # # Here, the server will yield the current stereo images that are set
        self.stereo_image = dexterity_pb2.StereoImage() # left_image = buffer.tobytes(), right_image = buffer2.tobytes())


    def start(self):
        self.server = grpc.server(futures.ThreadPoolExecutor(max_workers=8))
        dexterity_pb2_grpc.add_HandTrackingServiceServicer_to_server(
            self.HandTrackingServicer(self), self.server)
        self.server.add_insecure_port(f'[::]:{self.port}')
        self.server.start()
        print(f"Server started, listening on port {self.port}")
        threading.Thread(target=self._keep_alive, daemon=True).start()

    def stop(self):
        if self.server:
            self.server.stop(0)
            print("Server stopped")

    def _keep_alive(self):
        try:
            while True:
                time.sleep(86400)
        except KeyboardInterrupt:
            self.stop()

    def get_latest(self):
        return self.latest

    class HandTrackingServicer(dexterity_pb2_grpc.HandTrackingServiceServicer):
        def __init__(self, outer):
            print("HandTrackingServicer init")
            self.outer = outer


        def StreamHandUpdatesandGetSimStates(self, request_iterator, context):
            for hand_update in request_iterator:
                try: 
                    start_transformation_time = time.time()
                    # print(f"Transformation started at: {start_transformation_time} seconds")
                    transformations = {
                        "left_wrist": self.outer.axis_transform @ process_matrix(hand_update.left_hand.wristMatrix),
                        "right_wrist": self.outer.axis_transform @ process_matrix(hand_update.right_hand.wristMatrix),
                        "left_fingers": process_matrices(hand_update.left_hand.skeleton.jointMatrices),
                        "right_fingers": process_matrices(hand_update.right_hand.skeleton.jointMatrices),
                        "head": rotate_head(self.outer.axis_transform @ process_matrix(hand_update.Head)),
                        "left_pinch_distance": get_pinch_distance(hand_update.left_hand.skeleton.jointMatrices),
                        "right_pinch_distance": get_pinch_distance(hand_update.right_hand.skeleton.jointMatrices),
                    }
                    transformations["right_wrist_roll"] = get_wrist_roll(transformations["right_wrist"])
                    transformations["left_wrist_roll"] = get_wrist_roll(transformations["left_wrist"])

                    self.outer.latest = transformations
                    end_transformation_time = time.time()
                    # print(f"Transformation ended at: {end_transformation_time} seconds")
                    elapsed_time = end_transformation_time - start_transformation_time
                    # print(f"Elapsed time: {elapsed_time * 1000} ms")
                    
                    # You can add any processing here to fill sim_states based on transformations
                    yield self.outer.sim_states

                except Exception as e: 
                    print(e)


        def StreamHandUpdatesandGetStereoImages(self, request_iterator, context):
            for hand_update in request_iterator:
                try:

                    # print(f"Transformation started at: {start_transformation_time} seconds")
                    transformations = {
                        "left_wrist": self.outer.axis_transform @ process_matrix(hand_update.left_hand.wristMatrix),
                        "right_wrist": self.outer.axis_transform @ process_matrix(hand_update.right_hand.wristMatrix),
                        "left_fingers": process_matrices(hand_update.left_hand.skeleton.jointMatrices),
                        "right_fingers": process_matrices(hand_update.right_hand.skeleton.jointMatrices),
                        "head": rotate_head(self.outer.axis_transform @ process_matrix(hand_update.Head)  @ rot_x(60)),
                        "left_pinch_distance": get_pinch_distance(hand_update.left_hand.skeleton.jointMatrices),
                        "right_pinch_distance": get_pinch_distance(hand_update.right_hand.skeleton.jointMatrices),
                    }
                    transformations["right_wrist_roll"] = get_wrist_roll(transformations["right_wrist"])
                    transformations["left_wrist_roll"] = get_wrist_roll(transformations["left_wrist"])

                    self.outer.latest = transformations

                    # print("real_image:", image.shape)
                    # w = image.shape[1]//2
                    # h = image.shape[0]
                    # im_left = image[:, :w]
                    # im_right = image[:, w:2*w]
                    
                    # _, buffer = cv2.imencode('.jpg', im_left)
                    # _, buffer2 = cv2.imencode('.jpg', im_right)
                    # print("encoded both")
                    
                    # # Here, the server will yield the current stereo images that are set
                    # yield dexterity_pb2.StereoImage(left_image = buffer.tobytes(), right_image = buffer2.tobytes())

                    yield self.outer.stereo_image

                except Exception as e:
                    print(f"Error streaming stereo images: {e}")


        def Check(self, request, context):
            response = dexterity_pb2.HealthCheckResponse()
            response.status = dexterity_pb2.HealthCheckResponse.SERVING
            return response

    def set_sim_states(self, sim_dict):
        # You can add any processing here to set sim_states
        for key, value in sim_dict.items(): 
            if self.thirdperson: 
                self.sim_states.matrices[key].CopyFrom(matrix4x4_to_proto(ZUP2YUP @ self.convert @ value))
            else: 
                self.sim_states.matrices[key].CopyFrom(matrix4x4_to_proto(ZUP2YUP @ value))

    # Method to update stereo images, automatically applying JPEG encoding
    def set_image_states(self, stereo_image_np, stereo = False):
        # Extract the numpy arrays from the input dictionary
        image = stereo_image_np
        # convert rgb to bgr
        image = cv2.cvtColor(image, cv2.COLOR_RGB2BGR)

        if stereo: 
            w = image.shape[1] // 2
            h = image.shape[0]
            print("sim_image:", image.shape)

            # Split the image into left and right
            im_left = image[:,:w]
            im_right = image[:,w:2*w]

            _, buffer_left = cv2.imencode('.jpg', im_left)
            _, buffer_right = cv2.imencode('.jpg', im_right)


        else:

            # Encode both images as JPEG using OpenCV
            _, buffer_left = cv2.imencode('.jpg', image)
            _, buffer_right = cv2.imencode('.jpg', image)

        # Create and set the StereoImage message
        stereo_image = dexterity_pb2.StereoImage(left_image = buffer_left.tobytes(), right_image = buffer_right.tobytes())

        self.stereo_image = stereo_image

        # ret, image = self.cap.read() # reads frame that has stereo images side by side
        # if not ret:
        #     print("failed to grab")
        #     return 

        # # split image in half
        # image = cv2.rotate(image, cv2.ROTATE_180)
        # w = image.shape[1]//2
        # h = image.shape[0]
        # im_left = image[:, :w]
        # im_right = image[:, w:2*w]
        
        # _, buffer = cv2.imencode('.jpg', im_left)
        # _, buffer2 = cv2.imencode('.jpg', im_right)
        # print("encoded both")

        # self.stereo_image.left_image = buffer.tobytes()
        # self.stereo_image.right_image = buffer2.tobytes()
        # pass 



    def _encode_jpeg(self, image):
        """Encodes a PIL image to JPEG format."""
        try:
            output = io.BytesIO()
            image.save(output, format='JPEG')
            jpeg_data = output.getvalue()

            return jpeg_data
        
        except Exception as e:
            print(f"JPEG encoding error: {e}")
            return None



# Usage example
if __name__ == '__main__':


    server = HandTrackingServer()
    server.start()

    try:
        while True:
            latest = server.get_latest()
            if latest:
                pass
                # print(f"Latest head position: {latest['head'][0, :3, 3]}")
    except KeyboardInterrupt:
        server.stop()



from dm_control import mjcf
import dexhub 
import numpy as np 
import mujoco
from pathlib import Path
from scipy.spatial.transform import Rotation as R 
# from robot_descriptions import robotiq_2f85_mj_description, panda_mj_description, allegro_hand_mj_description, aloha_mj_description, ur5e_mj_description
import os 
from dart_physics.utils.mesh_download import * 
import subprocess 
from dart_physics.utils.robot_keyframe_names import * 
from copy import deepcopy 

_HERE = Path(__file__).parent

def obj2mjcf(obj_name, sdf = False ):

    try: 
        path = download_mesh(obj_name)
        obj_path = os.path.join(path, obj_name, f"{obj_name}.xml")
        print(obj_path)
        if os.path.exists(obj_path):
            print("Found cached convex decomposition!")
            return mjcf.from_path(obj_path)

    except: 

        xml_path = os.path.join(_HERE, "..",  "assets", "custom", obj_name, f"{obj_name}.xml")
        return mjcf.from_path(xml_path)
        
        
    
    print(path)
    if sdf: 
        process = subprocess.Popen(
            [
                "obj2mjcf",
                "--obj-dir",
                str(path),
                "--save-mjcf",
                "--verbose",
            ],
            stdin=subprocess.PIPE,
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
        )

    else: 
        process = subprocess.Popen(
            [
                "obj2mjcf",
                "--obj-dir",
                str(path),
                "--save-mjcf",
                "--decompose",
                "--verbose",
                # "--coacd-args.preprocess-resolution", 
                # "100", 
                # "--coacd-args.threshold", 
                # "0.01", 
                # "--coacd-args.mcts-iterations", 
                # "200", 
            ],
            stdin=subprocess.PIPE,
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
        )

    # Write 'n' to the process input
    stdout, stderr = process.communicate(input=b'n\n')

    return mjcf.from_path(obj_path)




def construct_scene(task_cfg, robot_cfg, thirdperson = False): 


    root = mjcf.RootElement()
    root.statistic.meansize = 0.08
    getattr(root.visual, "global").azimuth = -120
    getattr(root.visual, "global").elevation = -20

    root.option.timestep = ".002"
    root.option.integrator = task_cfg.get("integrator", "implicitfast")
    if robot_cfg["name"] == "DualPanda": 
        root.option.integrator = "implicitfast"
    # if robot_cfg["name"] == "RBY1": 
    #     root.option.integrator = "implicitfast"

    root.option.cone = robot_cfg.get("cone", "pyramidal")
    root.option.o_solref = task_cfg.get("o_solref", ".0000000001 1")
    root.option.o_solimp = task_cfg.get("o_solimp", "0.01 1")
    root.option.impratio = task_cfg.get("impratio", "10000.0")
    # root.option.o_friction= "1 1 0.5 0.5"
    root.option.flag.multiccd = task_cfg.get("override", "disable")
    root.option.flag.override = task_cfg.get("override", "enable")
    root.option.solver = task_cfg.get("solver", "Newton")
    # if robot_cfg["name"] == "RBY1": 
    #     root.option.solver = "PGS"

    root.size.memory = "1000M" 

    root.worldbody.add("light", pos="0 0 1.5", directional="true")

    root.asset.add("texture", name="grid", type="2d", builtin="checker", rgb1="0.2 0.3 0.4", rgb2="0.1 0.2 0.3", width="256", height="256", mark="cross", markrgb=".0 .0 .0")
    root.asset.add("material", name="grid", texture="grid", texrepeat="1 1", texuniform="true")

    floor = root.worldbody.add("geom", name="floor", size="5 5 0.01", type="plane", material="grid")

    table_height = task_cfg.get("table_height", 0.0)    
    

    for robot_name, robot in robot_cfg['robots'].items():
        robot: RobotConfig
        robot_mjcf = mjcf.from_path(robot.xml_path)
        robot_mjcf.model = robot_name

        if robot.freejoint: 
            root.attach(robot_mjcf).add("freejoint")

        elif robot.attach_to is not None: 

            pos = robot.attach_to[:3]
            quat = robot.attach_to[3:]

            site = root.worldbody.add("site", name=f"{robot_name}_site", pos=pos,  quat=quat)
            site.attach(robot_mjcf)            


        for new_body in robot.add_bodies: 
            new_body: NewBodyConfig
            add_new_body_to_model(root, robot_name, new_body.attach_to, \
                                  new_body.body_name, new_body.pos, new_body.quat)



    try: 
        eq = root.equality
    except:
        eq = root.add("equality")

    for finger in robot_cfg["reverse_points"].values() if thirdperson else robot_cfg["points"].values():

        add_mocap_body_to_model(root, finger.mocap_frame, 1)

        if finger.weld: 
            eq.add("weld", body1 = f"{finger.mocap_frame}", body2 = finger.body_frame) # solimp="1 1 0.5", solref="0.01 0.3")




    root, object_default_qpos = add_object_to_model(root, task_cfg)
    try: 
        for obj in task_cfg["objects"]:
            attach_viz_axis_to_body(root, obj + "/" + obj, [0, 0, 0], 6)
    except:
        pass 

    for obj in task_cfg["objects"]:
        try: 
            df = root.default.find("default", f"{obj}/collision")
            # dff = df.find("default", "cambridge_mug/collision")
            df.geom.condim = 4
            df.geom.friction = "10 10 10"
        except:
            print(f"Object condim / friction {obj} not found in the model")


    home_q = [] 
    for robot_name, robot in robot_cfg['robots'].items(): 
        if robot.home_q is not None: 
            home_q.append(np.array(robot.home_q))

    try: 
        qpos = np.concatenate(home_q + [np.concatenate(object_default_qpos)])
    except:
        qpos = np.concatenate(home_q)
    print(len(qpos))
    qpos_string = " ".join([str(q) for q in qpos])
    
    for robot_name in robot_cfg['robots'].keys(): 
        root.keyframe.remove(f"{robot_name}/home")

    root.keyframe.add("key", name="home", time=0, qpos= qpos_string)

    # add front camera 

    # cam_body = add_mocap_body_to_model(root, "main_front") 
    # cam_body.pos = [-0.036, 1.767, 0.819]
    # cam_body.quat = R.from_euler("xyz", [0, 0, 0], degrees = True).as_quat(scalar_first = True)

    root.worldbody.add("camera", name="main_front", fovy=70, pos=[-0.0, 0.8, 0.819], xyaxes=[-1.000, -0.00, -0.000, 0.008, -1.00, 1.0])
# <camera pos="-0.036 1.767 0.819" xyaxes="-1.000 -0.020 -0.000 0.008 -0.420 0.907"/>


    robot_cams = robot_cfg.get("cameras", [])

    for i, camera in enumerate(robot_cams):
        camera: CameraInfo 
        
        if camera.attach_to is not None: 
            assert camera.attach_type is not None 

            print(f"Attaching camera {i} to {camera.attach_to} of type {camera.attach_type}")

            mat = R.from_euler("xyz", camera.euler, degrees = False).as_matrix() 
            mat = mat @ rot_y(-90)[:3, :3] @ rot_z(-90)[:3, :3]
            quat = R.from_matrix(mat).as_quat(scalar_first = True)


            root, m = attach_body_to_body(root, body_name = camera.attach_to, mocap_name = camera.name, \
                                    pos = camera.pos, quat = quat, size = 1)
            m.add("camera", name=camera.name, fovy=camera.fovy)

        else: 

            cam_body = add_mocap_body_to_model(root, camera.name) 
            cam_body.pos = camera.pos
            cam_body.quat = R.from_euler("xyz", camera.euler, degrees = True).as_quat(scalar_first = True)

            cam_body.add("camera", name=camera.name, fovy=camera.fovy)


    if "lights" in task_cfg:
        for light_cfg in task_cfg["lights"]:
            light_cfg = deepcopy(light_cfg)
            del light_cfg["custom"]
            root.worldbody.add("light", **light_cfg)

    if "render" in task_cfg:
        width, height = task_cfg["render"]["resolution"]

    getattr(root.visual, "global").offwidth = 640
    getattr(root.visual, "global").offheight = 640

    # # Add cameras to the model
    if "render" in task_cfg:
        cameras_cfg = task_cfg["render"].get("cameras", [])
        cameras_body = add_mocap_body_to_model(root, "head2", len(cameras_cfg))
        
        for i, camera_cfg in enumerate(cameras_cfg):
            camera_cfg = deepcopy(camera_cfg)
            mocap_body_name = f"camera_{camera_cfg['name']}"
            cameras_body = add_mocap_body_to_model(root, mocap_body_name, size=1)
            
            # Remove pos & rot from the camera config
            # Position and rotation will be set by the mocap body
            camera_cfg["pos"] = [0, 0, 0]
            for key in ["quat", "eulers", "euler", "axisangle", "xyaxes", "zaxis"]:
                if key in camera_cfg:
                    del camera_cfg[key]
                
            # Add resolution to the camera
            camera_cfg["resolution"] = task_cfg["render"]["resolution"]
                
            # Add camera to the model
            # print(f"Adding camera {i} with config: {camera_cfg}")
            cameras_body.add("camera", **camera_cfg)

    try: 
        material_names = ["white", "off_white", "black", "green", "light_blue"]
        reflectance = 0.2
        shininess = reflectance * 2
        specular = reflectance * 0.5

        for robot in robot_cfg["robots"].keys():
            for material_name in material_names:
                root.asset.find("material", f"{robot}/{material_name}").reflectance = reflectance
                root.asset.find("material", f"{robot}/{material_name}").shininess = shininess
                root.asset.find("material", f"{robot}/{material_name}").specular = specular

        print("material property set.")

    except:
        pass 


    # save xml     
    with open("scene.xml", "w") as f:
        f.write(root.to_xml_string())

    # dexhub.register_sim(root)
    
    # if meshcat: 

    #     from utils.meshcat_viz import MJCMeshcat
    #     from io import BytesIO

    #     mjc = MJCMeshcat()

    #     meshcat_dict = {} 

    #     meshcat_dict["material"] = {}
    #     for material in root.asset.find_all("material"):
    #         name = material.full_identifier 
    #         rgba = material.rgba
    #         texture = material.texture

    #         if texture is not None and texture != "grid": 
    #             texture_contents = BytesIO(texture.file.contents)
    #         else: 
    #             texture_contents = None
            
    #         # for method in dir(material):
    #         #     print(method, ':', str(getattr(material, method)))  

    #         mjc.register_material(name, texture_contents, rgba)

    #     # while True:
    #     #     pass 
            

    #     for mesh in root.asset.find_all("mesh"):

    #         input_io = BytesIO(mesh.file.contents)

    #         try: 
    #             scale = mesh.scale[0]
    #         except:
    #             scale = 1.0

    #         # print(mesh.name, mesh.file.extension, scale)
    #         mjc.register_mesh(mesh.name, mesh.file.extension, input_io, scale)


    #     for body in root.worldbody.find_all("body"):
    #         if body.geom is None: 
    #             continue
    #         for geom in body.geom:
    #             # print(dir(geom))
    #             # print(geom.__class__)
    #             # for method in dir(geom):
    #             #     print(method, ':', str(getattr(geom, method)))
    #             if "collision" in str(geom.dclass): 
    #                 pass
    #             else:
    #                 try: 
    #                     if geom.material is None: 
    #                         # print(str(geom.dclass))
    #                         material = geom.dclass.geom.material 
    #                     else:
    #                         material = geom.material
    #                     mjc.register_geom(body.full_identifier, geom.mesh.name, material.full_identifier, geom.pos, geom.quat)
    #                 except Exception as e:
    #                     # print(body.full_identifier, geom.name, e)
    #                     pass 

    # else:
    #     mjc = None 

    return mujoco.MjModel.from_xml_string(root.to_xml_string(), root.get_assets()), root 


def add_new_body_to_model(root, robot_name, attach_to, body_name, pos_string, quat_string): 

    rf = root.find("body", f"{robot_name}/{attach_to}")
    rfb = rf.add("body", name=body_name, pos= pos_string, quat = quat_string)
    # rfb.add(
    #     "geom",
    #     group = 5, 
    #     type="sphere",
    #     size=".02 .02 .02",
    #     contype="0",
    #     conaffinity="0",
    #     rgba=".6 .3 .3 .7",
    # )




def add_mocap_body_to_model(root, name, size = 1, pos = [0, 0, 0], quat = [1, 0, 0, 0]): 
    default = 0.01 

    size = size * default

    pos = " ".join([str(p) for p in pos])
    quat = " ".join([str(q) for q in quat])

    site = root.worldbody.add("body", name=name, mocap=True)
    site.add(
        "geom",
        type="box",
        size=f"{size} {size} {size}",
        contype="0",
        conaffinity="0",
        rgba=".3 .6 .3 .9",
        group = 4, 
    )
    site.add(
        "geom",
        type="box",
        size=f"{size} .002 .002",
        contype="0",
        conaffinity="0",
        rgba="1 0 0 1",
        pos=f"{size} 0 0",
        group = 4, 
    )
    # length 0.07 cylinder with 0.01 radius in green color indicating y-axis
    site.add(
        "geom",
        type="box",
        size=f".002 {size} .002",
        contype="0",
        conaffinity="0",
        rgba="0 1 0 1",
        pos = f"0 {size} 0",
        group = 4, 

    )
    # length 0.07 cylinder with 0.01 radius in blue color indicating z-axis
    site.add(
        "geom",
        type="box",
        size=f".002 .002 {size}",
        contype="0",
        conaffinity="0",
        rgba="0 0 1 1",
        pos = f"0 0 {size}",
        group = 4, 

    )

    return site

def add_targets_to_model(root): 

    body = root.worldbody.add("body", name="l_target", mocap=True)
    body.add(
        "geom",
        group = 5, 
        type="box",
        size=".05 .05 .05",
        contype="0",
        conaffinity="0",
        rgba=".3 .6 .3 .5",
    )
    # length 0.07 cylinder with 0.01 radius in red color indicating x-axis
    body.add(
        "geom",
        type="box",
        size=".10 .007 .007",
        contype="0",
        conaffinity="0",
        rgba="1 0 0 1",
        pos=".05 0 0",
    )
    # length 0.07 cylinder with 0.01 radius in green color indicating y-axis
    body.add(
        "geom",
        type="box",
        size=".007 .10 .007",
        contype="0",
        conaffinity="0",
        rgba="0 1 0 1",
        pos = "0 .05 0"
    )
    # length 0.07 cylinder with 0.01 radius in blue color indicating z-axis
    body.add(
        "geom",
        type="box",
        size=".007 .007 .10",
        contype="0",
        conaffinity="0",
        rgba="0 0 1 1",
        pos = "0 0 .05"
    )

    body = root.worldbody.add("body", name="r_target", mocap=True)
    body.add(
        "geom",
        type="box",
        size=".05 .05 .05",
        contype="0",
        conaffinity="0",
        rgba=".3 .3 .6 .5",
    )
    body.add(
        "geom",
        type="box",
        size=".10 .007 .007",
        contype="0",
        conaffinity="0",
        rgba="1 0 0 1",
        pos=".05 0 0",
    )
    # length 0.07 cylinder with 0.01 radius in green color indicating y-axis
    body.add(
        "geom",
        type="box",
        size=".007 .10 .007",
        contype="0",
        conaffinity="0",
        rgba="0 1 0 1",
        pos = "0 .05 0"
    )
    # length 0.07 cylinder with 0.01 radius in blue color indicating z-axis
    body.add(
        "geom",
        type="box",
        size=".007 .007 .10",
        contype="0",
        conaffinity="0",
        rgba="0 0 1 1",
        pos = "0 0 .05"
    )

    return root 



def attach_viz_axis_to_body(root, body_name, pos, size = 1): 

    default = 0.01 

    size = size * default
    print(body_name)
    body = root.worldbody.find("body", body_name)
    pos_str = " ".join([str(p) for p in pos])
    if "/" in body_name:
        site = body.add("body", name=body_name.split('/')[1] + "_axis", pos=pos_str)
    else:
        site = body.add("body", name=body_name + "_axis", pos=pos_str)
    site.add(
        "geom",
        group = 5, 
        type="box",
        size=f"{size * 0.5} {size * 0.5} {size * 0.5}",
        contype="0",
        conaffinity="0",
        rgba=".6 .3 .3 .5",
    )
    # length 0.07 cylinder with 0.01 radius in red color indicating x-axis
    site.add(
        "geom",
        group = 5, 
        type="box",
        size=f"{size} .002 .002",
        contype="0",
        conaffinity="0",
        rgba="1 0 0 .5",
        pos=f"{size} 0 0",
    )
    # length 0.07 cylinder with 0.01 radius in green color indicating y-axis
    site.add(
        "geom",
        group = 5, 
        type="box",
        size=f".002 {size} .002",
        contype="0",
        conaffinity="0",
        rgba="0 1 0 .5",
        pos = f"0 {size} 0"
    )
    # length 0.07 cylinder with 0.01 radius in blue color indicating z-axis
    site.add(
        "geom",
        group = 5, 
        type="box",
        size=f".002 .002 {size}",
        contype="0",
        conaffinity="0",
        rgba="0 0 1 .5",
        pos = f"0 0 {size}", 
    )

    return root

def attach_body_to_body(root, body_name, mocap_name, pos, quat, size = 1): 

    default = 0.01 

    size = size * default
    print(body_name)
    body = root.worldbody.find("body", body_name)
    pos_str = " ".join([str(p) for p in pos])
    quat_str = " ".join([str(q) for q in quat])
    site = body.add("body", name=mocap_name, pos=pos_str, quat=quat_str)
    site.add(
        "geom",
        group = 5, 
        type="box",
        size=f"{size * 0.5} {size * 0.5} {size * 0.5}",
        contype="0",
        conaffinity="0",
        rgba=".6 .3 .3 .5",
    )
    # length 0.07 cylinder with 0.01 radius in red color indicating x-axis
    site.add(
        "geom",
        group = 5, 
        type="box",
        size=f"{size} .002 .002",
        contype="0",
        conaffinity="0",
        rgba="1 0 0 .5",
        pos=f"{size} 0 0",
    )
    # length 0.07 cylinder with 0.01 radius in green color indicating y-axis
    site.add(
        "geom",
        group = 5, 
        type="box",
        size=f".002 {size} .002",
        contype="0",
        conaffinity="0",
        rgba="0 1 0 .5",
        pos = f"0 {size} 0"
    )
    # length 0.07 cylinder with 0.01 radius in blue color indicating z-axis
    site.add(
        "geom",
        group = 5, 
        type="box",
        size=f".002 .002 {size}",
        contype="0",
        conaffinity="0",
        rgba="0 0 1 .5",
        pos = f"0 0 {size}", 
    )

    return root, site



def add_object_to_model(model, task_cfg): 

    print("fixed objects", task_cfg.get("fixed_objects", []))


    for object in task_cfg['objects']:
        obj = obj2mjcf(object)   # mjcf.from_path("/Users/yhpark/workspace/mj_aws/assets/custom/yellow_drill/yellow_drill/yellow_drill.xml")
        print("processing  " + object)
        if "piano" in object: 
            model.worldbody.attach(obj)

        elif object in task_cfg.get("fixed_objects", []):
            
            body = model.worldbody.attach(obj)
            body.pos = task_cfg['default_poses'][object][:3]
            body.quat = task_cfg['default_poses'][object][3:7]

        else:
            model.worldbody.attach(obj).add("freejoint")

    object_default_qpos = [] 
    for obj in task_cfg['objects']:
        print("adding default joint pose  " + obj)

        if obj == "cube": 
            print("here?")

            scrambled_pose = '3.1426 -3.12861 1.56976 -0.000984066 -4.71318 -4.71315 0.500066 0.499657 -0.49985 -0.500427 -5.4361e-06 -0.00105736 0.00700177 0.999975 -0.000159023 0.707166 -0.000551277 0.707047 0.499996 -0.500004 -0.499944 0.500056 -0.000206243 0.000330177 -0.707384 0.706829 -0.707495 -0.000335452 0.706718 -0.000199843 -0.707043 0.004306 0.707141 -0.00483592 -0.50299 -0.496992 -0.502969 -0.497014 -0.000732543 1.04854e-06 -0.707166 -0.707047 1 -6.94691e-05 -0.000515106 0.000382513 0.499994 0.501014 -0.500005 0.498985 -0.00464879 -0.707305 -0.706879 0.00452857 0.00437059 0.707365 -0.706822 0.00420462 -0.000935246 -0.000398716 -0.70222 0.711959 -0.500046 -0.501052 0.500044 -0.498856 0.000141309 3.32215e-05 0.999982 -0.00592773 0.00127499 -0.706838 -0.707374 0.000114238 -0.70796 0.00437499 -0.706218 0.00552882 0.000173516 0.000281969 -0.70708 -0.707134 0.000199487 0.706843 -0.70737 -0.000112591'

            scrambled_pose = list(map(float, scrambled_pose.split()))

            pose = [0, 0.3, 0.1] + [1, 0, 0 ,0] + \
                    scrambled_pose
            # pose = [] 

        elif obj == "piano": 

            pose_key = "-0.000169732 -0.000623174 -0.000169732 -0.000169732 -0.000623174 -0.000169732 -0.000623174 -0.000169732 -0.000169732 -0.000623174 -0.000169732 -0.000623174 -0.000169732 -0.000623174 -0.000169732 -0.000169732 -0.000623174 -0.000169732 -0.000623174 -0.000169732 -0.000169732 -0.000623174 -0.000169732 -0.000623174 -0.000169732 -0.000623174 -0.000169732 -0.000169732 -0.000623174 -0.000169732 -0.000623174 -0.000169732 -0.000169732 -0.000623174 -0.000169732 -0.000623174 -0.000169732 -0.000623174 -0.000169732 -0.000169732 -0.000623174 -0.000169732 -0.000623174 -0.000169732 -0.000169732 -0.000623174 -0.000169732 -0.000623174 -0.000169732 -0.000623174 -0.000169732 -0.000169732 -0.000623174 -0.000169732 -0.000623174 -0.000169732 -0.000169732 -0.000623174 -0.000169732 -0.000623174 -0.000169732 -0.000623174 -0.000169732 -0.000169732 -0.000623174 -0.000169732 -0.000623174 -0.000169732 -0.000169732 -0.000623174 -0.000169732 -0.000623174 -0.000169732 -0.000623174 -0.000169732 -0.000169732 -0.000623174 -0.000169732 -0.000623174 -0.000169732 -0.000169732 -0.000623174 -0.000169732 -0.000623174 -0.000169732 -0.000623174 -0.000169732 -0.000169732"
            # split the string by space and convert to float
            pose = [float(x) for x in pose_key.split(" ")]


        elif obj in task_cfg.get("fixed_objects", []):
            print(object + " is fixed")
            pose = [] 

        else: 
            pose = task_cfg['default_poses'][obj]

        object_default_qpos.append(np.array(pose))

    # b = model.worldbody.add("body")
    # b.add("geom", type="box", size=".05 .05 .05", rgba="1 0 0 1")
    # b.add("freejoint")
    # object_default_qpos.append(np.array([0.0, 0.0, 0.026, 0.0, 0.0, 0.0, 1.0]))

    return model, object_default_qpos



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: dexterity.proto
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import symbol_database as _symbol_database
from google.protobuf.internal import builder as _builder
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()




DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n\x0f\x64\x65xterity.proto\x12\tdexterity\"\xa7\x01\n\tMatrix4x4\x12\x0b\n\x03m00\x18\x01 \x01(\x02\x12\x0b\n\x03m01\x18\x02 \x01(\x02\x12\x0b\n\x03m02\x18\x03 \x01(\x02\x12\x0b\n\x03m03\x18\x04 \x01(\x02\x12\x0b\n\x03m10\x18\x05 \x01(\x02\x12\x0b\n\x03m11\x18\x06 \x01(\x02\x12\x0b\n\x03m12\x18\x07 \x01(\x02\x12\x0b\n\x03m13\x18\x08 \x01(\x02\x12\x0b\n\x03m20\x18\t \x01(\x02\x12\x0b\n\x03m21\x18\n \x01(\x02\x12\x0b\n\x03m22\x18\x0b \x01(\x02\x12\x0b\n\x03m23\x18\x0c \x01(\x02\"7\n\x08Skeleton\x12+\n\rjointMatrices\x18\x01 \x03(\x0b\x32\x14.dexterity.Matrix4x4\"X\n\x04Hand\x12)\n\x0bwristMatrix\x18\x01 \x01(\x0b\x32\x14.dexterity.Matrix4x4\x12%\n\x08skeleton\x18\x02 \x01(\x0b\x32\x13.dexterity.Skeleton\"y\n\nHandUpdate\x12\"\n\tleft_hand\x18\x01 \x01(\x0b\x32\x0f.dexterity.Hand\x12#\n\nright_hand\x18\x02 \x01(\x0b\x32\x0f.dexterity.Hand\x12\"\n\x04Head\x18\x03 \x01(\x0b\x32\x14.dexterity.Matrix4x4\"6\n\x0bStereoImage\x12\x12\n\nleft_image\x18\x01 \x01(\x0c\x12\x13\n\x0bright_image\x18\x02 \x01(\x0c\"\x88\x01\n\tSimStates\x12\x34\n\x08matrices\x18\x01 \x03(\x0b\x32\".dexterity.SimStates.MatricesEntry\x1a\x45\n\rMatricesEntry\x12\x0b\n\x03key\x18\x01 \x01(\t\x12#\n\x05value\x18\x02 \x01(\x0b\x32\x14.dexterity.Matrix4x4:\x02\x38\x01\"\x14\n\x12HealthCheckRequest\"\x8f\x01\n\x13HealthCheckResponse\x12<\n\x06status\x18\x01 \x01(\x0e\x32,.dexterity.HealthCheckResponse.ServingStatus\":\n\rServingStatus\x12\x0b\n\x07UNKNOWN\x10\x00\x12\x0b\n\x07SERVING\x10\x01\x12\x0f\n\x0bNOT_SERVING\x10\x02\x32\x90\x02\n\x13HandTrackingService\x12U\n StreamHandUpdatesandGetSimStates\x12\x15.dexterity.HandUpdate\x1a\x14.dexterity.SimStates\"\x00(\x01\x30\x01\x12Z\n#StreamHandUpdatesandGetStereoImages\x12\x15.dexterity.HandUpdate\x1a\x16.dexterity.StereoImage\"\x00(\x01\x30\x01\x12\x46\n\x05\x43heck\x12\x1d.dexterity.HealthCheckRequest\x1a\x1e.dexterity.HealthCheckResponseb\x06proto3')

_globals = globals()
_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, _globals)
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'dexterity_pb2', _globals)
if _descriptor._USE_C_DESCRIPTORS == False:

  DESCRIPTOR._options = None
  _SIMSTATES_MATRICESENTRY._options = None
  _SIMSTATES_MATRICESENTRY._serialized_options = b'8\001'
  _globals['_MATRIX4X4']._serialized_start=31
  _globals['_MATRIX4X4']._serialized_end=198
  _globals['_SKELETON']._serialized_start=200
  _globals['_SKELETON']._serialized_end=255
  _globals['_HAND']._serialized_start=257
  _globals['_HAND']._serialized_end=345
  _globals['_HANDUPDATE']._serialized_start=347
  _globals['_HANDUPDATE']._serialized_end=468
  _globals['_STEREOIMAGE']._serialized_start=470
  _globals['_STEREOIMAGE']._serialized_end=524
  _globals['_SIMSTATES']._serialized_start=527
  _globals['_SIMSTATES']._serialized_end=663
  _globals['_SIMSTATES_MATRICESENTRY']._serialized_start=594
  _globals['_SIMSTATES_MATRICESENTRY']._serialized_end=663
  _globals['_HEALTHCHECKREQUEST']._serialized_start=665
  _globals['_HEALTHCHECKREQUEST']._serialized_end=685
  _globals['_HEALTHCHECKRESPONSE']._serialized_start=688
  _globals['_HEALTHCHECKRESPONSE']._serialized_end=831
  _globals['_HEALTHCHECKRESPONSE_SERVINGSTATUS']._serialized_start=773
  _globals['_HEALTHCHECKRESPONSE_SERVINGSTATUS']._serialized_end=831
  _globals['_HANDTRACKINGSERVICE']._serialized_start=834
  _globals['_HANDTRACKINGSERVICE']._serialized_end=1106
# @@protoc_insertion_point(module_scope)



from .base import MatrixLieGroup
from .se3 import SE3
from .so3 import SO3
from .utils import get_epsilon, skew

__all__ = (
    "SE3",
    "SO3",
    "MatrixLieGroup",
    "get_epsilon",
    "skew",
)



import dart_physics.mink as mink
from typing import Dict 
import numpy as np 
from dart_physics.robot_cfgs import IKTasks, PostureTask, RelativeCoMTask, CoMTask, DampingTask

class DiffIK: 
    
    def __init__(self, configuration: mink.Configuration, robot_cfg: Dict, thirdperson = False): 

        self.configuration = configuration
        self.model = self.configuration.model
        self.data = self.configuration.data 

        self.thirdperson = thirdperson

        self.tasks = [] 

        self.regulation_tasks = [ ] 
        self.scales = [] 

        regulation_tasks = robot_cfg.get("regulation_task", [])

        for task in regulation_tasks: 
            # task: PostureTask or RelativeCoMTask or CoMTask
            if isinstance(task, PostureTask): 
                t: PostureTask
                t = mink.PostureTask(model=self.model, cost=task.cost)

                if task.qpos is not None: 
                    t.set_target(task.qpos)
                else:   
                    t.set_target_from_configuration(self.configuration)

                if task.disable_joints is not None: 
                    t.disable_joints(task.disable_joints)

                self.regulation_tasks.append(t)

            elif isinstance(task, RelativeCoMTask): 
                t = mink.RelativeComTask(root_name=task.root_name, root_type=task.root_type, cost=task.cost)
                t.set_target_from_configuration(self.configuration)
                self.regulation_tasks.append(t)
            elif isinstance(task, CoMTask): 
                t = mink.ComTask(cost=task.cost)
                t.set_target_from_configuration(self.configuration)
                self.regulation_tasks.append(t)

            elif isinstance(task, DampingTask): 
                t = mink.DampingTask(model=self.model, cost=task.cost)
                t.disable_joints(task.disable_joints)
                self.regulation_tasks.append(t)


        for task in robot_cfg['ik_task']: 
            task: IKTasks

            point = "points" if not self.thirdperson else "reverse_points"

            # if frame is tuple: 
            if task.type == "relative":

                root = robot_cfg[point][task.root]
                target = robot_cfg[point][task.target]

                pos_cost = task.pos_cost
                ori_cost = task.ori_cost

                self.tasks.append(mink.RelativeFrameTask(
                    root_name = root.body_frame,
                    root_type = root.type,
                    frame_name = target.body_frame,
                    frame_type = target.type,
                    position_cost = pos_cost,
                    orientation_cost = ori_cost
                ))
                self.scales.append(task.scale)

            elif task.type == "absolute":

                target = robot_cfg[point][task.target]
                
                pos_cost = task.pos_cost
                ori_cost = task.ori_cost

                self.tasks.append(mink.FrameTask(
                    frame_name = target.body_frame,
                    frame_type = "body",
                    position_cost = pos_cost,
                    orientation_cost = ori_cost,
                ))
                self.scales.append(task.scale)

        
        joint_names = robot_cfg["joints"]
        
        max_velocities = {joint: 10 * np.pi for joint in joint_names}

        self.solver = "quadprog"
        self.pos_threshold = 1e-4
        self.ori_threshold = 1e-4
        self.max_iters = 20

        self.limits = [
            mink.ConfigurationLimit(model = self.model, gain= 0.99,  min_distance_from_limits=0.0001),
            # velocity_limit,
            # mink.CollisionAvoidanceLimit(
            #     model=self.model,
            #     geom_pairs=collision_pairs,
            #     minimum_distance_from_collisions=0.01,
            #     collision_detection_distance=0.01,
            # ),
        ]



    def solve(self): 

        """
        input: targets containing dictionary of mocap frames
        """

        try:
            vel = mink.solve_ik(
                self.configuration, self.tasks + self.regulation_tasks, 0.002, self.solver, 1e-3, limits=self.limits,  
            )

            qtarget = self.configuration.integrate(vel, 0.002)

            return qtarget
        
        except Exception as e: 

            print(e)

            return self.configuration.data.qpos 

    def update(self, q): 
        self.configuration.update(q=q)



# utils/camera_viewer.py
import zmq
import numpy as np
import matplotlib.pyplot as plt
from queue import Queue
from threading import Thread
import time
import multiprocessing as mp

def data_receiver(socket, data_queue):
    while True:
        msg = socket.recv_pyobj()  # Receive the image data
        data_queue.put(msg)

def image_rendering_process(addr, stereo = False):
    context = zmq.Context()
    socket = context.socket(zmq.SUB)
    socket.connect(addr)
    socket.setsockopt(zmq.SUBSCRIBE, b'')

    data_queue = Queue()

    # Start the data receiving thread
    receiver_thread = Thread(target=data_receiver, args=(socket, data_queue))
    receiver_thread.daemon = True
    receiver_thread.start()

    plt.ion()  # Turn on interactive mode
    fig, ax = plt.subplots()

    # Initialize the image plot with some dummy data
    if stereo: 
        img = np.zeros((640, 1920, 3), dtype=np.uint8)  # Assuming a 480x640 RGB image
    else:
        img = np.zeros((640, 960, 3), dtype=np.uint8)
    im_plot = ax.imshow(img)

    plt.tight_layout()

    while True:
        while not data_queue.empty():
            img_data = data_queue.get()

            # Update the image plot with the new image data
            im_plot.set_data(img_data)

            ax.draw_artist(im_plot)
            fig.canvas.flush_events()  # Update the plot

        time.sleep(0.01)  # Small sleep to prevent overloading the CPU

def start_camera_viewer_process(addr, stereo = False):
    # Start the image rendering process using multiprocessing
    process = mp.Process(target=image_rendering_process, args=(addr,stereo, ))
    process.start()
    return process

def get_socket(addr):
    context = zmq.Context()
    socket = context.socket(zmq.PUB)
    socket.bind(addr)
    return socket


# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: handtracking.proto
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import symbol_database as _symbol_database
from google.protobuf.internal import builder as _builder
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()




DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n\x12handtracking.proto\x12\x0chandtracking\"\xdb\x01\n\tMatrix4x4\x12\x0b\n\x03m00\x18\x01 \x01(\x02\x12\x0b\n\x03m01\x18\x02 \x01(\x02\x12\x0b\n\x03m02\x18\x03 \x01(\x02\x12\x0b\n\x03m03\x18\x04 \x01(\x02\x12\x0b\n\x03m10\x18\x05 \x01(\x02\x12\x0b\n\x03m11\x18\x06 \x01(\x02\x12\x0b\n\x03m12\x18\x07 \x01(\x02\x12\x0b\n\x03m13\x18\x08 \x01(\x02\x12\x0b\n\x03m20\x18\t \x01(\x02\x12\x0b\n\x03m21\x18\n \x01(\x02\x12\x0b\n\x03m22\x18\x0b \x01(\x02\x12\x0b\n\x03m23\x18\x0c \x01(\x02\x12\x0b\n\x03m30\x18\r \x01(\x02\x12\x0b\n\x03m31\x18\x0e \x01(\x02\x12\x0b\n\x03m32\x18\x0f \x01(\x02\x12\x0b\n\x03m33\x18\x10 \x01(\x02\"\xb7\x01\n\x0e\x43ontrolCommand\x12\x39\n\x07\x63ommand\x18\x01 \x01(\x0e\x32(.handtracking.ControlCommand.CommandType\x12\x17\n\x0f\x61\x64\x64itional_info\x18\x02 \x01(\t\"Q\n\x0b\x43ommandType\x12\x13\n\x0fSTART_RECORDING\x10\x00\x12\x12\n\x0eSTOP_RECORDING\x10\x01\x12\x08\n\x04SAVE\x10\x02\x12\x0f\n\x0bRESET_SCENE\x10\x03\"3\n\x0f\x43ontrolResponse\x12\x0f\n\x07success\x18\x01 \x01(\x08\x12\x0f\n\x07message\x18\x02 \x01(\t\":\n\x08Skeleton\x12.\n\rjointMatrices\x18\x01 \x03(\x0b\x32\x17.handtracking.Matrix4x4\"^\n\x04Hand\x12,\n\x0bwristMatrix\x18\x01 \x01(\x0b\x32\x17.handtracking.Matrix4x4\x12(\n\x08skeleton\x18\x02 \x01(\x0b\x32\x16.handtracking.Skeleton\"\x82\x01\n\nHandUpdate\x12%\n\tleft_hand\x18\x01 \x01(\x0b\x32\x12.handtracking.Hand\x12&\n\nright_hand\x18\x02 \x01(\x0b\x32\x12.handtracking.Hand\x12%\n\x04Head\x18\x03 \x01(\x0b\x32\x17.handtracking.Matrix4x4\" \n\rHandUpdateAck\x12\x0f\n\x07message\x18\x01 \x01(\t\"$\n\x05\x46rame\x12\x0c\n\x04left\x18\x01 \x01(\x0c\x12\r\n\x05right\x18\x02 \x01(\x0c\"\x1c\n\tStreamAck\x12\x0f\n\x07message\x18\x01 \x01(\t\"\x8c\x01\n\x08Matrices\x12\x36\n\x08matrices\x18\x01 \x03(\x0b\x32$.handtracking.Matrices.MatricesEntry\x1aH\n\rMatricesEntry\x12\x0b\n\x03key\x18\x01 \x01(\t\x12&\n\x05value\x18\x02 \x01(\x0b\x32\x17.handtracking.Matrix4x4:\x02\x38\x01\x32\x62\n\x13HandTrackingService\x12K\n\x11StreamHandUpdates\x12\x18.handtracking.HandUpdate\x1a\x18.handtracking.HandUpdate\"\x00\x30\x01\x32N\n\x0bVideoStream\x12?\n\x0cSendMatrices\x12\x16.handtracking.Matrices\x1a\x17.handtracking.StreamAck2e\n\x0e\x43ontrolService\x12S\n\x12SendControlCommand\x12\x1c.handtracking.ControlCommand\x1a\x1d.handtracking.ControlResponse\"\x00\x62\x06proto3')

_globals = globals()
_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, _globals)
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'handtracking_pb2', _globals)
if _descriptor._USE_C_DESCRIPTORS == False:

  DESCRIPTOR._options = None
  _MATRICES_MATRICESENTRY._options = None
  _MATRICES_MATRICESENTRY._serialized_options = b'8\001'
  _globals['_MATRIX4X4']._serialized_start=37
  _globals['_MATRIX4X4']._serialized_end=256
  _globals['_CONTROLCOMMAND']._serialized_start=259
  _globals['_CONTROLCOMMAND']._serialized_end=442
  _globals['_CONTROLCOMMAND_COMMANDTYPE']._serialized_start=361
  _globals['_CONTROLCOMMAND_COMMANDTYPE']._serialized_end=442
  _globals['_CONTROLRESPONSE']._serialized_start=444
  _globals['_CONTROLRESPONSE']._serialized_end=495
  _globals['_SKELETON']._serialized_start=497
  _globals['_SKELETON']._serialized_end=555
  _globals['_HAND']._serialized_start=557
  _globals['_HAND']._serialized_end=651
  _globals['_HANDUPDATE']._serialized_start=654
  _globals['_HANDUPDATE']._serialized_end=784
  _globals['_HANDUPDATEACK']._serialized_start=786
  _globals['_HANDUPDATEACK']._serialized_end=818
  _globals['_FRAME']._serialized_start=820
  _globals['_FRAME']._serialized_end=856
  _globals['_STREAMACK']._serialized_start=858
  _globals['_STREAMACK']._serialized_end=886
  _globals['_MATRICES']._serialized_start=889
  _globals['_MATRICES']._serialized_end=1029
  _globals['_MATRICES_MATRICESENTRY']._serialized_start=957
  _globals['_MATRICES_MATRICESENTRY']._serialized_end=1029
  _globals['_HANDTRACKINGSERVICE']._serialized_start=1031
  _globals['_HANDTRACKINGSERVICE']._serialized_end=1129
  _globals['_VIDEOSTREAM']._serialized_start=1131
  _globals['_VIDEOSTREAM']._serialized_end=1209
  _globals['_CONTROLSERVICE']._serialized_start=1211
  _globals['_CONTROLSERVICE']._serialized_end=1312
# @@protoc_insertion_point(module_scope)



from pathlib import Path
import os

CURR_PATH = Path(__file__).resolve().parent

def get_config_path(task_name: str) -> str:
    """
    Get path to config file for task.
    """
    return os.path.join(CURR_PATH, f"{task_name}.yaml")


import mujoco

SUPPORTED_FRAMES = ("body", "geom", "site")

FRAME_TO_ENUM = {
    "body": mujoco.mjtObj.mjOBJ_BODY,
    "geom": mujoco.mjtObj.mjOBJ_GEOM,
    "site": mujoco.mjtObj.mjOBJ_SITE,
}
FRAME_TO_JAC_FUNC = {
    "body": mujoco.mj_jacBody,
    "geom": mujoco.mj_jacGeom,
    "site": mujoco.mj_jacSite,
}
FRAME_TO_POS_ATTR = {
    "body": "xpos",
    "geom": "geom_xpos",
    "site": "site_xpos",
}
FRAME_TO_XMAT_ATTR = {
    "body": "xmat",
    "geom": "geom_xmat",
    "site": "site_xmat",
}


def dof_width(joint_type: mujoco.mjtJoint) -> int:
    """Get the dimensionality of the joint in qvel."""
    return {0: 6, 1: 3, 2: 1, 3: 1}[joint_type]


def qpos_width(joint_type: mujoco.mjtJoint) -> int:
    """Get the dimensionality of the joint in qpos."""
    return {0: 7, 1: 4, 2: 1, 3: 1}[joint_type]



"""Joint velocity limit."""

from typing import Mapping

import mujoco
import numpy as np
import numpy.typing as npt

from ..configuration import Configuration
from ..constants import dof_width
from .exceptions import LimitDefinitionError
from .limit import Constraint, Limit


class VelocityLimit(Limit):
    """Inequality constraint on joint velocities in a robot model.

    Floating base joints are ignored.

    Attributes:
        indices: Tangent indices corresponding to velocity-limited joints.
        limit: Maximum allowed velocity magnitude for velocity-limited joints, in
            [m]/[s] for slide joints and [rad]/[s] for hinge joints.
        projection_matrix: Projection from tangent space to subspace with
            velocity-limited joints.
    """

    indices: np.ndarray
    limit: np.ndarray
    projection_matrix: np.ndarray

    def __init__(
        self,
        model: mujoco.MjModel,
        velocities: Mapping[str, npt.ArrayLike] = {},
    ):
        """Initialize velocity limits.

        Args:
            model: MuJoCo model.
            velocities: Dictionary mapping joint name to maximum allowed magnitude in
                [m]/[s] for slide joints and [rad]/[s] for hinge joints.
        """
        limit_list: list[float] = []
        index_list: list[int] = []
        for joint_name, max_vel in velocities.items():
            jid = model.joint(joint_name).id
            jnt_type = model.jnt_type[jid]
            jnt_dim = dof_width(jnt_type)
            jnt_id = model.jnt_dofadr[jid]
            if jnt_type == mujoco.mjtJoint.mjJNT_FREE:
                raise LimitDefinitionError(f"Free joint {joint_name} is not supported")
            max_vel = np.atleast_1d(max_vel)
            if max_vel.shape != (jnt_dim,):
                raise LimitDefinitionError(
                    f"Joint {joint_name} must have a limit of shape ({jnt_dim},). "
                    f"Got: {max_vel.shape}"
                )
            index_list.extend(range(jnt_id, jnt_id + jnt_dim))
            limit_list.extend(max_vel.tolist())

        self.indices = np.array(index_list)
        self.indices.setflags(write=False)
        self.limit = np.array(limit_list)
        self.limit.setflags(write=False)

        dim = len(self.indices)
        self.projection_matrix = np.eye(model.nv)[self.indices] if dim > 0 else None

    def compute_qp_inequalities(
        self, configuration: Configuration, dt: float
    ) -> Constraint:
        r"""Compute the configuration-dependent joint velocity limits.

        The limits are defined as:

        .. math::

            -v_{\text{max}} \cdot dt \leq \Delta q \leq v_{\text{max}} \cdot dt

        where :math:`v_{max} \in {\cal T}` is the robot's velocity limit
        vector and :math:`\Delta q \in T_q({\cal C})` is the displacement in the
        tangent space at :math:`q`. See the :ref:`derivations` section for
        more information.

        Args:
            configuration: Robot configuration :math:`q`.
            dt: Integration timestep in [s].

        Returns:
            Pair :math:`(G, h)` representing the inequality constraint as
            :math:`G \Delta q \leq h`, or ``None`` if there is no limit.
        """
        del configuration  # Unused.
        if self.projection_matrix is None:
            return Constraint()
        G = np.vstack([self.projection_matrix, -self.projection_matrix])
        h = np.hstack([dt * self.limit, dt * self.limit])
        return Constraint(G=G, h=h)



"""Exceptions specific to mink."""

from typing import Sequence

import mujoco


class MinkError(Exception):
    """Base class for Mink exceptions."""


class UnsupportedFrame(MinkError):
    """Exception raised when a frame type is unsupported."""

    def __init__(self, frame_type: str, supported_types: Sequence[str]):
        message = (
            f"{frame_type} is not supported."
            f"Supported frame types are: {supported_types}"
        )
        super().__init__(message)


class InvalidFrame(MinkError):
    """Exception raised when a frame name is not found in the robot model."""

    def __init__(
        self,
        frame_name: str,
        frame_type: str,
        model: mujoco.MjModel,
    ):
        if frame_type == "body":
            available_names_of_type_frame_type = [
                model.body(i).name for i in range(model.nbody)
            ]
        elif frame_type == "site":
            available_names_of_type_frame_type = [
                model.site(i).name for i in range(model.nsite)
            ]
        else:
            assert frame_type == "geom"
            available_names_of_type_frame_type = [
                model.geom(i).name for i in range(model.ngeom)
            ]

        message = (
            f"{frame_type} '{frame_name}' does not exist in the model. "
            f"Available {frame_type} names: {available_names_of_type_frame_type}"
        )

        super().__init__(message)


class InvalidKeyframe(MinkError):
    """Exception raised when a keyframe name is not found in the robot model."""

    def __init__(self, keyframe_name: str, model: mujoco.MjModel):
        available_keyframes = [model.key(i).name for i in range(model.nkey)]
        message = (
            f"Keyframe {keyframe_name} does not exist in the model. "
            f"Available keyframe names: {available_keyframes}"
        )
        super().__init__(message)


class InvalidMocapBody(MinkError):
    """Exception raised when a body is not a mocap body."""

    def __init__(self, mocap_name: str, model: mujoco.MjModel):
        available_mocap_names = [
            model.body(i).name
            for i in range(model.nbody)
            if model.body(i).mocapid[0] != -1
        ]
        message = (
            f"Body '{mocap_name}' is not a mocap body. "
            f"Available mocap bodies: {available_mocap_names}"
        )
        super().__init__(message)


class NotWithinConfigurationLimits(MinkError):
    """Exception raised when a configuration violates its limits."""

    def __init__(
        self,
        joint_id: int,
        value: float,
        lower: float,
        upper: float,
        model: mujoco.MjModel,
    ):
        joint_name = model.joint(joint_id).name
        message = (
            f"Joint {joint_id} ({joint_name}) violates configuration limits "
            f"{lower} <= {value} <= {upper}"
        )
        super().__init__(message)



import torch 
from avp_stream3.utils.se3_utils import * 


VISIONOS_TO_ISAAC = torch.tensor([[1, 0, 0, 0], 
                                  [0, 0, -1, 0], 
                                  [0, 1, 0, 0],
                                  [0, 0, 0, 1]], dtype = torch.float32).unsqueeze(0)

FRAME_TRANSFORMATION = torch.tensor([[-1, 0, 0],
                                    [0, 0, 1],
                                    [0, 1, 0]], dtype = torch.float32).unsqueeze(0)
FRAME_QUAT = isaac_mat2quat(FRAME_TRANSFORMATION)

FRAME_44 = torch.tensor([[-1, 0, 0, 0.7],
                         [0, 0,  1, 0.9],
                         [0, 1,  0, 1.6],
                         [0, 0, 0, 1]], dtype = torch.float32).unsqueeze(0)


ROTATE_90DEG_AROUND_X = torch.tensor([[1, 0, 0, 0],
                            [0, 0, -1, 0],
                            [0, 1, 0, 0],
                            [0, 0, 0, 1]], dtype = torch.float32).unsqueeze(0)

ROTATE_90DEG_AROUND_Y = torch.tensor([[0, 0, 1, 0],
                            [0, 1, 0, 0],
                            [-1, 0, 0, 0],
                            [0, 0, 0, 1]], dtype = torch.float32).unsqueeze(0)

ROTATE_NEG_90DEG_AROUND_Y = torch.tensor([[0, 0, -1, 0],
                            [0, 1, 0, 0],
                            [1, 0, 0, 0],
                            [0, 0, 0, 1]], dtype = torch.float32).unsqueeze(0)

ROTATE_180DEG_AROUND_Z = torch.tensor([[-1, 0, 0, 0],
                                       [0, -1, 0, 0],
                                       [0, 0, 1, 0],
                                       [0, 0, 0, 1]], dtype=torch.float32).unsqueeze(0)


VISIONOS_RIGHT_HAND_TO_LEAP = ROTATE_90DEG_AROUND_X @ ROTATE_NEG_90DEG_AROUND_Y

VISIONOS_LEFT_HAND_TO_LEAP = ROTATE_90DEG_AROUND_X @ ROTATE_90DEG_AROUND_Y

VISIONOS_RIGHT_FINGERS_TO_LEAP = ROTATE_90DEG_AROUND_Y



import pyspacemouse
import time
import numpy as np
import zmq

def main():     

    r_device = pyspacemouse.open(dof_callback=pyspacemouse.silent_callback, button_callback=pyspacemouse.silent_callback, DeviceNumber = 0) 
    l_device = pyspacemouse.open(dof_callback=pyspacemouse.silent_callback, button_callback=pyspacemouse.silent_callback, DeviceNumber = 2) 
    if r_device:
        # while True: 
        while True: 
            r_state, l_state = r_device.read(), l_device.read()
            # print xyz of right and left spacemouse with 3 decimals 
            # state.x, state.y, state.z 
            # in one line, use format 

            print("[L] x: {:+.3f}, y: {:+.3f}, z: {:+.3f} roll: {:+.3f}, pitch: {:+.3f}, yaw: {:+.3f}"\
                  .format(l_state.x, l_state.y, l_state.z, l_state.roll, l_state.pitch, l_state.yaw) + "       " + \
                  "[R] x: {:+.3f}, y: {:+.3f}, z: {:+.3f} roll: {:+.3f}, pitch: {:+.3f}, yaw: {:+.3f}"\
                  .format(r_state.x, r_state.y, r_state.z, r_state.roll, r_state.pitch, r_state.yaw), end="\r")
            
            time.sleep(0.01)



if __name__ == "__main__": 

    main()


import numpy as np 
from dart_physics.utils.run_functions import *


class CameraInfo: 

    def __init__(self, name, pos = [0, 0, 0], euler = [0, 0, 0], fovy = 47.5, attach_to = None, attach_type = None): 
    
        self.name = name 
        self.attach_to = attach_to
        self.attached_robot, self.attached_body = attach_to.split("/") if attach_to is not None else None, None
        self.attach_type = attach_type
        self.pos = pos
        self.fovy = fovy 
        self.euler = euler


    def __repr__(self):
            
        return f"CameraInfo(name={self.name}, attach_to={self.attach_to}, pos={self.pos}, quat={self.quat}, fovy={self.fovy}, euler={self.euler})"
    

class BodyInfo: 

    def __init__(self, name, geoms, pos= [0,0,0], quat = [1, 0, 0, 0]): 

        self.name = name 
        self.geoms = geoms

        if isinstance(pos, str): 
            pos = [float(p) for p in pos.split()]
        if isinstance(quat, str):
            quat = [float(q) for q in quat.split()]

        self.pos = pos
        self.quat = quat

    def __repr__(self):
            
        return f"BodyInfo(geoms={self.geoms}, pos={self.pos}, quat={self.quat})"

class GeomInfo: 

    def __init__(self, mesh, pos= [ 0, 0, 0] , quat = [1, 0, 0, 0]): 

        self.mesh = mesh

        if isinstance(pos, str):
            pos = [float(p) for p in pos.split()]
        if isinstance(quat, str):
            quat = [float(q) for q in quat.split()]

        self.pos = pos 
        self.quat = quat 

    def __repr__(self): 

        return f"GeomInfo(mesh={self.mesh}, pos={self.pos}, quat={self.quat})"

class FingerKeypoints:
    def __init__(self, chilarity, avp_idx, mocap_frame, body_frame, weld = False, type = "body", avp_transform = np.eye(3), init_posquat = None):
        self.chilarity = chilarity
        self.avp_idx = avp_idx
        self.mocap_frame = mocap_frame
        self.body_frame = body_frame
        self.weld = weld 
        self.type = type 
        self.avp_transform = avp_transform
        self.init_posquat = init_posquat

    def __repr__(self):
        return f"Finger(avp_idx={self.avp_idx}, mocap_frame={self.mocap_frame}, body_frame={self.body_frame})"

class IKTasks: 

    def __init__(self, target, root = None, pos_cost = 1.0, ori_cost = 0.0, scale = 1.0): 

        self.target = target 
        self.root = root 
        self.scale = scale
        self.pos_cost = pos_cost    
        self.ori_cost = ori_cost

        if root is None: 
            self.type = "absolute"
        else:
            self.type = "relative"

    def __repr__(self): 

        return f"IKTasks(target={self.target}, root={self.root}, scale={self.scale})"
    
class PostureTask: 

    def __init__(self, qpos, cost, disable_joints): 
        self.qpos = qpos
        self.cost = cost
        self.disable_joints = disable_joints

    def __repr__(self): 
        return f"PostureTask(qpos={self.qpos}, cost={self.cost}, disable_joints={self.disable_joints})"


class CoMTask: 

    def __init__(self, cost): 
        self.cost = cost

    def __repr__(self): 
        return f"CoMTask(cost={self.cost})"
    

class RelativeCoMTask: 

    def __init__(self, root_name, root_type, cost): 
        self.root_name = root_name
        self.root_type = root_type
        self.cost = cost

    def __repr__(self): 
        return f"RelativeCoMTask(root_name={self.root_name}, root_type={self.root_type}, cost={self.cost})"
    
class DampingTask: 

    def __init__(self, cost, disable_joints = None): 
        self.cost = cost
        self.disable_joints = disable_joints

    def __repr__(self): 
        return f"DampingTask(cost={self.cost}, disable_joints={self.disable_joints})"

class RobotConfig: 

    def __init__(self, xml_path, home_q, attach_to = None, freejoint = False, parallel_jaw = None, add_bodies = [] ): 

        self.home_q = home_q
        self.xml_path = xml_path
        self.freejoint = freejoint
        self.attach_to = attach_to
        self.parallel_jaw = parallel_jaw
        self.add_bodies = add_bodies

    def __repr__(self):

        return f"RobotConfig(home_q={self.home_q}, xml_path={self.xml_path})"


class NewBodyConfig: 

    def __init__(self, attach_to, body_name, pos, quat = [1, 0, 0, 0]): 

        self.attach_to = attach_to
        self.body_name = body_name
        self.pos = pos
        self.quat = quat

    def __repr__(self): 

        return f"NewBodyConfig(attach_to={self.attach_to}, body_name={self.body_name}, pos={self.pos}, quat={self.quat})"